    
from traits.api import \
    HasTraits, List, Array, Property, cached_property, \
    Instance, Trait, Button, on_trait_change, Tuple, \
    Int, Float, String

from traits.api import Enum

from traitsui.api import \
    View, Item

from ibvpy.core.i_sdomain import \
    ISDomain
    
from ibvpy.core.sdomain import \
    SDomain

from numpy import \
    array, unique, min, max, mgrid, ogrid, c_, alltrue, repeat, ix_, \
    arange, ones, zeros, multiply, sort, index_exp

from ibvpy.plugins.mayavi_util.pipelines import \
    MVPolyData, MVPointLabels, MVStructuredGrid


from .mcell import MGridCellSpec, MGridCell
from .elem_array_view import ElemArrayView
from functools import reduce


class MGridDomain(SDomain):
    '''
    Get an array with element node coordinates
    '''

    # Grid cell template - gets repeated according to the 
    # within the grid geometry specification
    #
    grid_cell = Property( depends_on = 'grid_cell_spec' )
    @cached_property
    def _get_grid_cell(self):
        return MGridCell( grid_cell_spec = self.grid_cell_spec )


    # Grid geometry specification
    #
    coord_min = Array( Float, value = [ 0., 0., 0.] )
    coord_max = Array( Float, value = [ 1., 1., 1.] )
    
    grid_cell_spec = Instance( MGridCellSpec )
    def _grid_cell_spec_default(self):
        return MGridCellSpec()
    
    # Remark[rch]: 
    # beware - the Int type is not regarded as a normal int
    # within an array and must be first converted to int array
    #
    # Had we defined int as the dtype of an array, there would
    # be errors during editing
    #
    shape   = Array( Int, value = [ 1, 1, 1 ] )
    
    # Derived specifier for element grid shape
    # It converts the Int array to int so that it can be
    # used by general numpy operators
    #
    egrid_shape = Property( depends_on = 'shape' )
    @cached_property
    def _get_egrid_shape(self):
        return self.shape.astype(int)

    #-------------------------------------------------------------------------
    # Shaping and slicing methods for construction and orientation
    #-------------------------------------------------------------------------
    def _get_grid_shape(self):
        '''
        Get the grid shape for the full index and point grids 
        '''
        cell_shape = self.grid_cell_spec.get_cell_shape().astype(int)
        egrid_shape = self.egrid_shape
        return multiply( cell_shape - 1, egrid_shape  ) + 1

    def _get_grid_size(self):
        '''Get the size of the full index and point grids
        '''
        shape = self._get_grid_shape()
        return reduce( lambda i,j: i*j, shape )
    
    def _get_igrid_slices(self):
        '''Get slices defining the index grid
        '''
        sub_cell_shape = self.grid_cell_spec.get_cell_shape() - 1
        elem_grid_shape = self.egrid_shape
        return tuple( [ slice( 0, c*g+1 ) 
                        for c,g in zip(sub_cell_shape,elem_grid_shape) ] )

    def _get_icell_slices(self):
        '''Get slices extracting the first cell from the index grid
        '''
        cell_shape = self.grid_cell_spec.get_cell_shape()
        return tuple( [ slice( 0,c )
                        for c in cell_shape ] )        

    def _get_icell(self):
        '''Get the node map within a cell of a 1-3 dimensional grid
        
        The enumeration of nodes within a single cell is managed by the
        self.grid_cell. This must be adapted to the global enumeration of the grid.
        
        The innermost index runs over the z-axis. Thus, the index of the points
        on the z axis is [0,1,...,shape[2]-1]. The following node at the y axis has 
        the number [shape[2], shape[2]+1, ..., shape[2]*2].

        '''
        igrid = arange( self._get_grid_size() ).reshape( self._get_grid_shape( ))
        # extract the first cell from the igrid
        icell_slices = self._get_icell_slices()
        icell = igrid[ icell_slices ]
        return icell

    #-------------------------------------------------------------------------
    # Generation methods for geometry and index maps
    #-------------------------------------------------------------------------

    def _get_pgrid_slices(self):
        '''Get the slices to be used for the mgrid tool 
        to generate the point grid.
        '''
        ndims = self.grid_cell_spec.get_n_dims()
        shape = self._get_grid_shape()
        return tuple( [ slice( float(self.coord_min[i]),
                               float(self.coord_max[i]),
                               complex(0,shape[i]) ) 
                        for i in range(ndims ) ] )

    pgrid = Property( depends_on = 'grid_cell_spec.+,shape,coord_min,coord_max' )
    @cached_property
    def _get_pgrid(self):
        '''
        Construct the point grid underlying the mesh grid structure.
        '''
        return mgrid[ self._get_pgrid_slices() ]

    def _get_pgrid_dims(self):
        return self._get_grid_shape()

    def _get_points(self):
        return c_[ tuple([ x.flatten() for x in self.pgrid ]) ]

    def _get_tpoints(self):
        '''Get the points in with deepest index along the x axes.
        '''
        tp = self.pgrid.swapaxes(1,3)
        return c_[ tuple([ x.flatten() for x in tp ]) ]
      
      
        
    # Base node grid
    #
    def _get_base_node_array(self):
        '''
        Construct the base node grid.
        '''
        # get the cell shape - number of cell points without the next base point
        subcell_shape = self.grid_cell_spec.get_cell_shape().astype(int) - 1
        # get the element grid shape (number of elements in each dimension)
        egrid_shape = self.egrid_shape

        # get the index offset between two neighboring points on the x-axis  
        z_offset = subcell_shape[2]
        # get the index offset between two neighboring points on the y-axis  
        y_offset = ( egrid_shape[2] * subcell_shape[2] + 1) * subcell_shape[1]
        # get the index offset between two neighboring points on the z-axis  
        x_offset = ( egrid_shape[2] * subcell_shape[2] + 1 ) * \
                   ( egrid_shape[1] * subcell_shape[1] + 1 ) * \
                     subcell_shape[0]

        # grid shape (shape + 1)
        gshape = egrid_shape + 1
        
        # generate points along an axis respecting the above offsets 
        xi_offsets = x_offset * arange( gshape[0] )
        yi_offsets = y_offset * arange( gshape[1] )
        zi_offsets = z_offset * arange( gshape[2] )

        # expand the dimension of offsets and sum them up. using broadcasting
        # this generates the grid of the base nodes (the last node is cut away 
        # as it does not hold any element
        igrid = xi_offsets[:-1,None,None] + \
                yi_offsets[None,:-1,None] + \
                zi_offsets[None,None,:-1]

        # return the indexes of the base nodes
        return sort( igrid.flatten() )

    def _get_elnode_map(self):
        '''
        Construct an array with 
        '''
        icell = self._get_icell()
        node_map   = icell.flatten()[ self.grid_cell.node_map ]
        base_nodes = self._get_base_node_array()

        # Use broadcasting to construct the node map for all elements
        #
        elnode_map = base_nodes[:,None] + node_map[None,:]
        return elnode_map

    def _get_expanded_elem_coords(self):
        iexp = index_exp[ self._get_elnode_map() ]
        return self._get_points()[iexp]

    def _get_elem_coords(self, elnum):
        iexp = index_exp[ self._get_elnode_map()[ elnum ] ]
        return self._get_points()[iexp]

    def _get_epoints(self):
        '''Get element node numbers
        
        Extract the indices of all points that are connected to an element.
        '''
        return self._get_points()[ ix_( self._get_epoint_numbers( ) ) ]
        
    def _get_epoint_numbers(self):
        # flatten and unique
        elnodes = self._get_elnode_map()
        return unique( elnodes.flatten() )

    show_elem_array = Button
    def _show_elem_array_fired(self):
        elem_array = self._get_elnode_map()
        self.show_array = ElemArrayView( data = elem_array, 
                                         rt_domain = self )
        self.show_array.configure_traits( kind = 'live' )
        
    #-----------------------------------------------------------------
    # Visualization related methods
    #-----------------------------------------------------------------

    mvp_pgrid = Trait( MVStructuredGrid )
    def _mvp_pgrid_default(self):
        return MVStructuredGrid( name = 'Point grid', 
                                  dims = self._get_pgrid_dims,
                                  points = self._get_tpoints )
    
    mvp_mgrid_ngeo_labels = Trait( MVPointLabels )
    def _mvp_mgrid_ngeo_labels_default(self):
        return MVPointLabels( name = 'Geo node numbers', 
                                  points = self._get_epoints,
                                  scalars = self._get_epoint_numbers)


    

    refresh_button = Button('Draw')
    @on_trait_change('refresh_button')
    def redraw(self):
        '''
        '''
        self.mvp_pgrid.redraw( )        
        #self.mvp_mgrid_ngeo_labels.redraw( 'label_scalars' )        

    #------------------------------------------------------------------
    # UI - related methods
    #------------------------------------------------------------------
    traits_view = View(Item('grid_cell_spec'),
                       Item('shape@'),
                       Item('coord_min'),
                       Item('coord_max'),
                       Item('refresh_button'),
                       Item('show_elem_array'),
                       resizable = True,
                       scrollable = True,
                       height = 0.5,
                       width = 0.5)        
    
if __name__ == '__main__':
    mgd = MGridDomain( shape = (2,1,1) )
    
    from ibvpy.plugins.ibvpy_app import IBVPyApp
    ibvpy_app = IBVPyApp( ibv_resource = mgd )
    ibvpy_app.main()
    
    
    


'''

@author: rch
'''


from traits.api import \
    HasStrictTraits, Instance, Button, Event, \
    DelegatesTo
from traits.etsconfig.api import ETSConfig
from traitsui.api import \
    TreeEditor, TreeNode, View, Item, VGroup, \
    ToolBar, \
    HSplit
from traitsui.menu import \
    Menu, MenuBar, Separator
from ibvpy.view.ui.bmcs_tree_node import \
    BMCSRootNode, BMCSTreeNode, BMCSLeafNode

import traits.api as tr

from .bmcs_study import BMCSStudy
from .bmcs_tree_view_handler import \
    menu_tools_report_pdf, menu_tools_report_tex, \
    BMCSTreeViewHandler, plot_self, menu_save, \
    menu_open, menu_exit, \
    toolbar_actions, key_bindings
from .i_bmcs_model import IBMCSModel


if ETSConfig.toolkit == 'wx':
    from traitsui.wx.tree_editor import \
        DeleteAction
if ETSConfig.toolkit == 'qt4':
    from traitsui.qt4.tree_editor import \
        DeleteAction
else:
    raise ImportError("tree actions for %s toolkit not available" %
                      ETSConfig.toolkit)


tree_node = TreeNode(node_for=[BMCSRootNode, BMCSTreeNode],
                     auto_open=False,
                     children='tree_node_list',
                     label='node_name',
                     view='tree_view',
                     menu=Menu(plot_self, DeleteAction),
                     )

leaf_node = TreeNode(node_for=[BMCSLeafNode],
                     auto_open=True,
                     children='',
                     label='node_name',
                     view='tree_view',
                     menu=Menu(plot_self)
                     )

tree_editor = TreeEditor(
    nodes=[tree_node, leaf_node],
    selected='selected_node',
    orientation='vertical'
)


class BMCSWindow(BMCSStudy):

    selected_node = Instance(HasStrictTraits)

    def _selected_node_changed(self):
        self.selected_node.ui = self

    def get_vot_range(self):
        return self.viz_sheet.get_vot_range()

    vot = DelegatesTo('viz_sheet')

    data_changed = Event

    replot = Button

    def _replot_fired(self):
        self.figure.clear()
        self.selected_node.plot(self.figure)
        self.data_changed = True

    clear = Button()

    def _clear_fired(self):
        self.figure.clear()
        self.data_changed = True

    view = View(
        HSplit(
            VGroup(
                Item('model',
                     id='bmcs.hsplit.left.tree.id',
                     dock='tab',
                     editor=tree_editor,
                     resizable=True,
                     show_label=False,
                     width=300,
                     height=200,
                     ),
                #                Item('selected_node@'),
                id='bmcs.hsplit.left.id',
                dock='tab',
            ),
            VGroup(
                Item('viz_sheet@',
                     show_label=False,
                     id='bmcs.hsplit.viz3d.notebook.id',
                     dock='tab',
                     ),
                dock='tab',
                id='bmcs.hsplit.viz3d.id',
                label='viz sheet',
            ),
            dock='tab',
            id='bmcs.hsplit.id',
        ),
        #        dock='tab',
        id='bmcs.id',
        width=1.0,
        height=1.0,
        title='BMCS',
        resizable=True,
        handler=BMCSTreeViewHandler(),
        key_bindings=key_bindings,
        toolbar=ToolBar(*toolbar_actions,
                        image_size=(32, 32),
                        show_tool_names=False,
                        show_divider=True,
                        name='view_toolbar'),
        menubar=MenuBar(Menu(menu_exit, Separator(),
                             menu_save, menu_open,
                             name='File'),
                        Menu(menu_tools_report_tex,
                             menu_tools_report_pdf,
                             name='Tools'),

                        )
    )


if __name__ == '__main__':
    from ibvpy.view.examples.response_tracer import ResponseTracer
    from ibvpy.sim.sim_base import Simulator
    from ibvpy.core.bcond_mngr import BCondMngr
    from ibvpy.bcond import BCDof, BCSlice
    bc_mngr = BCondMngr()
    bc_mngr.bcond_list = [
        BCDof(),
        BCSlice()]
    rt = ResponseTracer()
    bm = Simulator(node_name='sim',
                   tree_node_list=[BMCSTreeNode(node_name='subnode 1'),
                                   BMCSTreeNode(node_name='subnode 2'),
                                   rt,
                                   bc_mngr
                                   ])
    bm.tree_node_list += [bm.tline]
    tv = BMCSWindow(model=bm)
    rt.add_viz2d('time_profile', 'time profile')
    tv.configure_traits()



from scipy.optimize.zeros import brentq

from bmcs.time_functions import \
    LoadingScenario
from ibvpy.bcond import BCDof
from ibvpy.tmodel.mats3D.mats3D_plastic.vmats3D_desmorat import \
    MATS3DDesmorat
from simulator.api import Simulator, TLoop

from .interaction_scripts import run_rerun_test


class BrentqTLoop(TLoop):

    def eval(self):
        t_n1 = self.tline.val
        t_max = self.tline.max
        dt = self.tline.step
        while t_n1 <= t_max:
            print('\ttime: %g' % t_n1, end='')
            k = 0
            self.tstep.t_n1 = t_n1
            # run the iteration loop
            brentq
            while (k < self.k_max) and not self.user_wants_abort:
                if self.tstep.R_norm < self.acc:
                    print('\titer: %g' % k)
                    break
                self.tstep.make_iter()
                k += 1
            else:  # handle unfinished iteration loop
                if k >= self.k_max:  # add step size reduction
                    # no success abort the simulation
                    self.restart = True
                    print('Warning: '
                          'convergence not reached in %g iterations' % k)
                return
            # accept the time step and record the state in history
            self.tstep.make_incr()
            # update the line - launches notifiers to subscribers
            self.tline.val = t_n1
            # set a new target time
            t_n1 += dt
        return


class MATS3DDesmoratGrad(MATS3DDesmorat):
    tloop_type = BrentqTLoop


s = Simulator(
    model=MATS3DDesmorat()
)
bc = BCDof(
    var='u', dof=0, value=-0.001,
    time_function=LoadingScenario()
)
s.tstep.bcond_mngr.bcond_list = [bc]
run_rerun_test(s)



'''Example calculating the 3 point bending test 
with just one halve of the specimen included.


Hierarchy of configurable operators to perform an increment 
and iteration step

1. TimeBC - Time dependent boundary conditions 
2. DomainState - multiple uniform domains with state force and stiffness 
2.1 LocalGlobalMapping 
2.1.1 Kinematics - supplies the B matrix
      Can be parameterized at several leels - propose a structure
2.1.2 Discretization
2.2 TModel

Complying structure for response tracing
extracting data from several domains
'''

import time

from ibvpy.bcond import BCSlice
from ibvpy.fets import FETS2D4Q
from ibvpy.tmodel.mats2D import \
    MATS2DScalarDamage
from ibvpy.tmodel.viz3d_scalar_field import \
    Vis3DStateField, Viz3DScalarField
from mayavi import mlab
from simulator.api import \
    TStepBC, XDomainFEGrid


L = 600.0
H = 100.0
L_c = 5.0
a = 5.0
w_max = 0.4
dgrid1 = XDomainFEGrid(coord_max=(L, H),
                       shape=(20, 20),
                       integ_factor=50,
                       fets=FETS2D4Q())
x_x, x_y = dgrid1.mesh.geo_grid.point_x_grid
L_1 = x_x[1, 0]
d_L = L_c - L_1
x_x[1:, :] += d_L * (L - x_x[1:, :]) / (L - L_1)
a_L = a / H
n_a = int(a_L * dgrid1.shape[1])
fixed_right_bc = BCSlice(slice=dgrid1.mesh[-1, 0, -1, 0],
                         var='u', dims=[1], value=0)
fixed_x = BCSlice(slice=dgrid1.mesh[0, n_a:, 0, -1],
                  var='u', dims=[0], value=0)
control_bc = BCSlice(slice=dgrid1.mesh[0, -1, :, -1],
                     var='u', dims=[1], value=-w_max)
m = TStepBC(
    domains=[(dgrid1, MATS2DScalarDamage(algorithmic=True))],
    bc=[fixed_right_bc, fixed_x, control_bc],
    record={
        'damage': Vis3DStateField(var='omega'),
    }
)
s = m.sim
s.tloop.trait_set(verbose=True, k_max=200)
s.tline.step = 0.05
s.run()
time.sleep(3)
damage_viz = Viz3DScalarField(vis3d=s.hist['damage'])
damage_viz.setup()
damage_viz.plot(0.0)
mlab.show()




#-- Imports --------------------------------------------------------------------

from os.path \
    import join, dirname
    
from numpy \
    import sqrt
    
from numpy.random \
    import random

from traits.api \
    import HasTraits, Property, Array, Any, Event, \
    on_trait_change, Instance, WeakRef, Int, Str, Bool, Trait
    
from traitsui.api \
    import View, Item, TabularEditor, HSplit, Group
    
from traitsui.menu \
    import NoButtons, CancelButton
    
from traitsui.tabular_adapter \
    import TabularAdapter

from pyface.image_resource \
    import ImageResource
    
from ibvpy.view.plot3d.mayavi_util.pipelines import \
    MVPolyData, MVPointLabels, MVStructuredGrid    

#-- Tabular Adapter Definition -------------------------------------------------

class CoordArrayAdapter ( TabularAdapter ):

    columns = Property
    def _get_columns(self):
        return [('node', 'index'),
                ('x', 0),
                ('y', 1),
                ('z', 2) ]
        
    font        = 'Courier 10'
    alignment   = 'right'
    format      = '%g'
    index_text  = Property
    
    def _get_index_text ( self ):
        return str( self.row )
        
#-- Tabular Editor Definition --------------------------------------------------

coord_array_editor = TabularEditor(
    adapter = CoordArrayAdapter( ),
)

class ElemView(HasTraits):
    '''Get the element numbers.
    '''
    elem_num = Int(-1)
    elem_coords = Array
    view = View( Item( 'elem_num', style = 'readonly' ),
                 Item('elem_coords', editor = coord_array_editor, style = 'readonly' )
                 )
    

#-- Tabular Adapter Definition -------------------------------------------------

class ElemArrayAdapter ( TabularAdapter ):

    columns = Property
    def _get_columns(self):
        data = getattr( self.object, self.name )
        if len( data.shape ) != 2:
            raise ValueError('element node array must be two-dimensional')
        n_columns = getattr( self.object, self.name ).shape[1]

        cols = [ (str(i), i ) for i in range( n_columns ) ]
        return [ ('element', 'index') ] + cols
        
    font        = 'Courier 10'
    alignment   = 'right'
    format      = '%d'
    index_text  = Property
#    index_image = Property
    
    def _get_index_text ( self ):
        return str( self.row )
        
    def x_get_index_image ( self ):
        x, y, z = self.item
        if sqrt( (x - 0.5) ** 2 + (y - 0.5) ** 2 + (z - 0.5) ** 2 ) <= 0.25:
            return 'red_flag'
        return None

#-- Tabular Editor Definition --------------------------------------------------

elem_array_editor = TabularEditor(
    adapter = ElemArrayAdapter(),
    selected_row = 'current_row',
)

#-- ElemArrayView Class Definition -------------------------------------------------

class ElemArrayView ( HasTraits ):

    data = Array
    rt_domain = WeakRef
    
    
    mvp_elem_labels = Trait( MVPointLabels )
    def _mvp_elem_labels_default(self):
        return MVPointLabels( name = 'Geo node numbers', 
                                  points = self._get_current_elem_coords,
                                  scalars = self._get_current_elem_numbers,
                                  #color = (0.254902,0.411765,0.882353)
                                  color = (0.15,0.85,0.45))
                                 
        
    mvp_elem_geo = Trait( MVPolyData )
    def _mvp_elem_geo_default(self):
        return MVPolyData( name = 'Geo node numbers', 
                               points = self._get_current_elem_coords,
                               lines = self._get_current_elem_lines,
                               #color = (0.254902,0.411765,0.882353))
                               color= (0,55,0,75,0.0))
        
    
    show_elem = Bool(True)
            
    def _get_current_elem_coords(self):
        return self.rt_domain._get_elem_coords( self.current_row )
    
    def _get_current_elem_numbers(self):
        return self.data[self.current_row]
    
    def _get_current_elem_lines(self):
        line_map = self.rt_domain.grid_cell_spec.cell_lines
        return line_map
    
    current_row = Int(-1)
    @on_trait_change('current_row')
    def redraw(self):
        if self.show_elem:
            self.mvp_elem_labels.redraw('label_scalars')
            self.mvp_elem_geo.redraw()
    
    elem_view = Instance( ElemView )
    def _elem_view_default(self):
        return ElemView()
    
    @on_trait_change('current_row')
    def _display_current_row(self):
        if self.current_row != -1:
            self.elem_view.elem_num = self.current_row
            elem_coords = self.rt_domain._get_elem_coords( self.current_row )
            self.elem_view.elem_coords = elem_coords
    
    view = View(
                HSplit(
                       Item( 'data', editor = elem_array_editor, 
                             show_label = False, style = 'readonly' ),
                        Group(Item( 'elem_view@', show_label = False ),
                              Item( 'show_elem' , label = 'SHOW')
                             )),
        title     = 'Array Viewer',
        width     = 0.6,
        height    = 0.4,
        resizable = True,
        buttons   = [CancelButton]
    )
    
# Run the demo (if invoked from the command line):
if __name__ == '__main__':

    from traits.api import Button
    from .mdomain import MGridDomain

    rt_domain = MGridDomain( shape = (20,20,1) )
    demo = ElemArrayView( data = rt_domain._get_elnode_map(),
                          rt_domain = rt_domain  )
    demo.configure_traits()
    



'''Example calculating the 3 point bending test 
with just one halve of the specimen included.


Hierarchy of configurable operators to perform an increment 
and iteration step

1. TimeBC - Time dependent boundary conditions 
2. DomainState - multiple uniform domains with state force and stiffness 
2.1 LocalGlobalMapping 
2.1.1 Kinematics - supplies the B matrix
      Can be parameterized at several leels - propose a structure
2.1.2 Discretization
2.2 TModel

Complying structure for response tracing
extracting data from several domains
'''

import time

from ibvpy.bcond import BCSlice
from ibvpy.fets import FETS3D8H
from ibvpy.tmodel.mats2D import \
    MATS2DScalarDamage
from ibvpy.tmodel.mats3D.mats3D_microplane.vmats3D_mpl_d_eeq import \
    MATS3DMplDamageEEQ
from ibvpy.tmodel.mats3D.mats3D_plastic.vmats3D_desmorat import \
    MATS3DDesmorat
from ibvpy.tmodel.viz3d_scalar_field import \
    Vis3DStateField, Viz3DScalarField
from ibvpy.tmodel.viz3d_tensor_field import \
    Vis3DTensorField, Viz3DTensorField
from mayavi import mlab
from simulator.api import \
    TStepBC, XDomainFEGrid

from .mlab_decorators import decorate_figure


L = 600.0
H = 100.0
B = 50.0
L_c = 5.0
a = 5.0
w_max = 1
dgrid1 = XDomainFEGrid(dim_u=3,
                       coord_max=(L, H, B),
                       shape=(20, 5, 1),
                       fets=FETS3D8H())
x_x, x_y, y_z = dgrid1.mesh.geo_grid.point_x_grid
L_1 = x_x[1, 0]
d_L = L_c - L_1
x_x[1:, :] += d_L * (L - x_x[1:, :]) / (L - L_1)
a_H = a / H
n_a = int(a_H * dgrid1.shape[1])
fixed_right_bc = BCSlice(slice=dgrid1.mesh[-1, 0, -1, 0, :, :],
                         var='u', dims=[1], value=0)
fixed_x = BCSlice(slice=dgrid1.mesh[0, n_a:, :, 0, :, :],
                  var='u', dims=[0], value=0)
control_bc = BCSlice(slice=dgrid1.mesh[0, -1, :, 0, -1, :],
                     var='u', dims=[1], value=-w_max)

m = MATS3DDesmorat(tau_bar=6.0,
                   E_1=16000,
                   E_2=19000,
                   S=470e-6,
                   K=1300,
                   gamma=1100)

m_mic = MATS3DMplDamageEEQ(
    E=27000,
    nu=0.2,
    epsilon_0=59.0e-6,
    epsilon_f=250.0e-7,
    c_T=0.01
)

model = TStepBC(
    domains=[(dgrid1, m_mic)],
    bc=[fixed_right_bc, fixed_x, control_bc],
    record={
        'strain': Vis3DTensorField(var='eps_ab'),
        'stress': Vis3DTensorField(var='sig_ab'),
        #        'damage': Vis3DStateField(var='omega_a'),
    }
)
s = model.sim
s.tloop.k_max = 1000
s.tloop.verbose = True
s.tline.step = 0.05
print(model.hist)
print(model.hist.record_dict)
print(model.hist.record_dict['strain'])
s.run_thread()
time.sleep(8)

mlab.options.backend = 'envisage'

f_strain = mlab.figure()
scene = mlab.get_engine().scenes[-1]
scene.name = 'stress'
strain_viz = Viz3DTensorField(vis3d=model.hist['strain'])
strain_viz.setup()
strain_viz.warp_vector.filter.scale_factor = 100.0
strain_viz.plot(model.t_n)

f_stress = mlab.figure()
scene = mlab.get_engine().scenes[-1]
scene.name = 'stress'
stress_viz = Viz3DTensorField(vis3d=model.hist['stress'])
stress_viz.setup()
stress_viz.warp_vector.filter.scale_factor = 100.0
stress_viz.plot(model.t_n)


# f_damage = mlab.figure()
# scene = mlab.get_engine().scenes[-1]
# scene.name = 'damage'
#
# damage_viz = Viz3DScalarField(vis3d=s.hist['damage'])
# damage_viz.setup()
# damage_viz.lut_manager.use_default_range = True
# damage_viz.warp_vector.filter.scale_factor = 10.0
# damage_viz.plot(s.tstep.t_n)
# damage_viz.plot(0.0)


decorate_figure(f_stress, stress_viz, 800, [300, 40, 0])
decorate_figure(f_strain, strain_viz, 800, [300, 40, 0])
#decorate_figure(f_damage, damage_viz, 800, [300, 40, 0])

mlab.show()



'''This example couples two domains via 
an zero-thickness interface. 

Todo - simplify the domain-staet-xmodel-mesh hierarchy

The dependencies - Simulator - who sets the type of the time stepping
loop and the type of the time step.

Test two independent domains.
'''

import time

from ibvpy.bcond import BCSlice
from ibvpy.fets import FETS2D4Q
from ibvpy.fets.fets1D5 import FETS1D52ULRH
from ibvpy.tmodel.mats1D5.vmats1D5_e import MATS1D5Elastic
from ibvpy.tmodel.mats2D import \
    MATS2DScalarDamage
from ibvpy.tmodel.mats3D.mats3D_plastic.vmats3D_desmorat import \
    MATS3DDesmorat
from ibvpy.tmodel.viz3d_scalar_field import \
    Vis3DStateField, Viz3DScalarField
from ibvpy.tmodel.viz3d_tensor_field import \
    Vis3DTensorField, Viz3DTensorField
from mayavi import mlab
from simulator.api import \
    TStepBC
from simulator.xdomain.xdomain_fe_grid import XDomainFEGrid
from simulator.xdomain.xdomain_interface import XDomainFEInterface

import numpy as np


#from bmcs.tmodel.fets1d52ulrhfatigue import FETS1D52ULRHFatigue
n_x_e = 20
xdomain1 = XDomainFEGrid(coord_min=(0, 0),
                         coord_max=(100, 10),
                         shape=(n_x_e, 3),
                         fets=FETS2D4Q())
xdomain2 = XDomainFEGrid(coord_min=(0, 10),
                         coord_max=(100, 20),
                         shape=(n_x_e, 3),
                         fets=FETS2D4Q())
xdomain12 = XDomainFEInterface(
    I=xdomain1.mesh.I[:, -1],
    J=xdomain2.mesh.I[:, 0],
    fets=FETS1D52ULRH()
)

left_y = BCSlice(slice=xdomain1.mesh[0, 0, 0, 0],
                 var='u', dims=[1], value=0)
left_x = BCSlice(slice=xdomain1.mesh[0, :, 0, :],
                 var='u', dims=[0], value=-0)
right_x = BCSlice(slice=xdomain1.mesh[-1, :, -1, :],
                  var='u', dims=[0], value=0.1)
bc1 = [left_y, left_x, right_x]

m = TStepBC(
    domains=[(xdomain1, MATS2DScalarDamage(algorithmic=False)),
             (xdomain2, MATS2DScalarDamage(algorithmic=False)),
             #             (xdomain3, MATS3DDesmorat()),
             (xdomain12, MATS1D5Elastic(E_s=100, E_n=1)),
             #             (xdomain13, MATS1D5Elastic()),
             ],
    bc=bc1,  # + bc2,
    record={
        'strain': Vis3DTensorField(var='eps_ab'),
        #'damage': Vis3DStateField(var='omega'),
        #        'kinematic hardening': Vis3DStateField(var='z_a')
    }
)
s = m.sim
m.fe_domain[0].tmodel.omega_fn.f_t = 100.0
s.tloop.verbose = True
s.tloop.k_max = 1000
s.tline.step = 0.05
s.tstep.fe_domain.serialized_subdomains

xdomain12.hidden = True
s.run()
time.sleep(5)

mlab.options.backend = 'envisage'
f_strain = mlab.figure()
scene = mlab.get_engine().scenes[-1]
scene.name = 'strain'
strain_viz = Viz3DTensorField(vis3d=s.hist['strain'])
strain_viz.setup()
strain_viz.warp_vector.filter.scale_factor = 100.0
strain_viz.plot(s.tstep.t_n)

if False:
    f_damage = mlab.figure()
    scene = mlab.get_engine().scenes[-1]
    scene.name = 'damage'
    damage_viz = Viz3DScalarField(vis3d=s.hist['damage'])
    damage_viz.setup()
    damage_viz.warp_vector.filter.scale_factor = 100.0
    damage_viz.plot(s.tstep.t_n)


def decorate_figure(f, viz):
    mlab.view(0, 0, 140,
              np.array([50., 5.,  0.]), figure=f)
    mlab.orientation_axes(viz.src, figure=f)
    axes = mlab.axes(viz.src, figure=f)
    axes.label_text_property.trait_set(
        font_family='times', italic=False
    )
    axes.title_text_property.font_family = 'times'
    axes.axes.trait_set(
        x_label='x', y_label='y', z_label='z'
    )


#decorate_figure(f_damage, damage_viz)
decorate_figure(f_strain, strain_viz)
mlab.show()



'''
Created on 14. 4. 2014

@author: Vancikv
'''

import pickle

from ibvpy.util import \
    get_outfile
from pyface.api import ImageResource
from traits.api import \
    Button, Instance, WeakRef, Str
from traitsui.api import \
    View, Item, HGroup, Handler, \
    UIInfo, spring, VGroup, Label
from traitsui.file_dialog import \
    open_file, save_file
from traitsui.key_bindings import \
    KeyBinding, KeyBindings
from traitsui.menu import \
    Action


plot_self = Action(name='Plot', action='plot_node')
'''Context menu action for plotting tree nodes
'''
new_material = Action(name='New material', action='new_material')
'''Context menu action for adding a new database member
'''
del_material = Action(name='Delete', action='del_material')
'''Context menu action for deleting a database member
'''
menu_save = Action(name='Save', action='menu_save')
'''Menubar action for saving the root node to file
'''
menu_open = Action(name='Open', action='menu_open')
'''Menubar action for loading root node from file
'''
menu_exit = Action(name='Exit', action='menu_exit')
'''Menubar action for terminating the view
'''
menu_tools_report_tex = Action(name='Report as LaTeX source',
                               action='menu_tools_report_tex')
'''Menubar action for generation of report in LaTeX format 
'''

menu_tools_report_pdf = Action(name='Report as PDF document',
                               action='menu_tools_report_pdf')
'''Menubar action for generation of report in LaTeX format 
'''

key_bindings = KeyBindings(
    KeyBinding(binding1='Ctrl-r',
               description='Run simulation',
               method_name='run_action'),
    KeyBinding(binding1='Ctrl-p',
               description='Pause calculation',
               method_name='pause_action'),
    KeyBinding(binding1='Ctrl-s',
               description='Stop calculation',
               method_name='stop_action')
)


class BMCSTreeViewHandler(Handler):

    '''Handler for BMCSTreeView class
    '''
    # The UIInfo object associated with the view:
    info = Instance(UIInfo)
    node = WeakRef

    ok = Button('OK')
    cancel = Button('Cancel')
    delete = Button('OK')

    exit_view = View(
        VGroup(
            Label('Do you really wish to end '
                  'the session? Any unsaved data '
                  'will be lost.'),
            HGroup(Item('ok', show_label=False, springy=True),
                   Item('cancel', show_label=False, springy=True)
                   ),
        ),
        title='Exit dialog',
        kind='live'
    )

    del_view = View(
        HGroup(
            spring,
            Item(name='', label='Really delete?'),
            Item('delete', show_label=False),
            Item('cancel', show_label=False),
        ),
        kind='popup'
    )

    def new_material(self, info, node):
        mat_name = 'new_material_1'
        count = 1
        while node.get(mat_name, None):
            count += 1
            mat_name = 'new_material_' + str(count)
        node[mat_name] = node.klass()

    def del_material(self, info, node):
        if info.initialized:
            self.node = node
            self._ui = self.edit_traits(view='del_view')

    def plot_node(self, info, node):
        '''Handles context menu action Plot for tree nodes
        '''
        info.object.figure.clear()
        node.plot(info.object.figure)
        info.object.data_changed = True

    def menu_save(self, info):
        file_name = get_outfile(folder_name='.bmcs', file_name='')
        file_ = save_file(file_name=file_name)
        if file_:
            pickle.dump(info.object.root, open(file_, 'wb'), 1)

    def menu_open(self, info):
        file_name = get_outfile(folder_name='.bmcs', file_name='')
        file_ = open_file(file_name=file_name)
        if file_:
            info.object.root = pickle.load(open(file_, 'rb'))

    def menu_exit(self, info):
        if info.initialized:
            self.info = info
            self._ui = self.edit_traits(view='exit_view')

    def menu_tools_report_tex(self, info):
        info.object.report_tex()

    def menu_tools_report_pdf(self, info):
        info.object.report_pdf()

    def _delete_fired(self):
        del self.node.db[self.node.key]
        self._ui.dispose()

    def _ok_fired(self):
        self._ui.dispose()
        self.info.ui.dispose()

    def _cancel_fired(self):
        self._ui.dispose()

    #=========================================================================
    # Toolbar actions
    #=========================================================================

    def run_action(self, info):
        #print('Running action')
        info.object.run()

    def pause_action(self, info):
        #print('Running pause')
        info.object.pause()

    def stop_action(self, info):
        #print('Running stop')
        info.object.stop()

    def replot(self, info):
        #print('Running continue')
        info.object.replot()

    def clear(self, info):
        info.object.clear()

    def anim(self, info):
        info.object.anim()

    def render(self, info):
        info.object.render()

    def save(self, info):
        info.object.save()

    def load(self, info):
        info.object.load()


action_strings = \
    [('Plot', 'replot', 'Replot current diagrams'),
     ('Clear', 'clear', 'Clear current diagrams'),
     ('Save', 'save', 'Save session'),
     ('Load', 'load', 'Load session'),
     ('Animate', 'anim', 'Animate current session'),
     ('Render', 'render', 'Render current session')]

toolbar_actions = [Action(name="Run",
                          tooltip='Start computation',
                          enabled_when='enable_run',
                          image=ImageResource('kt-start'),
                          action="run_action"),
                   Action(name="Pause",
                          tooltip='Pause computation',
                          enabled_when='enable_pause',
                          image=ImageResource('kt-pause'),
                          action="pause_action"),
                   Action(name="Stop",
                          tooltip='Stop computation',
                          enabled_when='enable_stop',
                          image=ImageResource('kt-stop'),
                          action="stop_action"),
                   ]

# toolbar_actions += [Action(name=name,
#                            action=action,
#                            tooltip=tooltip)
#                     for name, action, tooltip in action_strings]



import time

from ibvpy.bcond import BCSlice
from ibvpy.fets import FETS2D4Q
from ibvpy.tmodel.mats2D.mats2D_elastic import MATS2DElastic
from ibvpy.tmodel.mats3D.mats3D_plastic.vmats3D_desmorat import \
    MATS3DDesmorat
from ibvpy.tmodel.viz3d_scalar_field import \
    Vis3DStateField, Viz3DScalarField
from ibvpy.tmodel.viz3d_tensor_field import \
    Vis3DTensorField, Viz3DTensorField
from mayavi import mlab
from simulator.api import \
    TStepBC
from simulator.xdomain.xdomain_fe_grid import XDomainFEGrid

import numpy as np

from .mlab_decorators import decorate_figure


thickness = 1

xdomain = XDomainFEGrid(coord_max=(1, 1),
                        shape=(1, 1),
                        integ_factor=thickness,
                        fets=FETS2D4Q())

# print(xmodel.B1_Eimabc.shape)
# print(xmodel.B0_Eimabc.shape)

m = MATS3DDesmorat()
m = MATS2DElastic(E=1, nu=0)

left_y = BCSlice(slice=xdomain.mesh[0, :, 0, :],
                 var='u', dims=[1], value=0)
left_x = BCSlice(slice=xdomain.mesh[0, :, 0, :],
                 var='u', dims=[0], value=-1)
right_x = BCSlice(slice=xdomain.mesh[-1, :, -1, :],
                  var='u', dims=[0], value=0.0)

m = TStepBC(
    domains=[(xdomain, m)],
    bc=[left_x, right_x, left_y],
    record={
        'strain': Vis3DTensorField(var='eps_ab'),
        #        'damage': Vis3DStateField(var='omega_a'),
        #        'kinematic hardening': Vis3DStateField(var='z_a')
    }
)
s = m.sim
s.tloop.k_max = 1000
s.tline.step = 0.1
s.tloop.verbose = True
s.run()

print('area', thickness)
F_ti = s.hist.F_t
print('left')
print(np.sum(F_ti[-1, right_x.dofs]))
print('right')
print(np.sum(F_ti[-1, left_x.dofs]))

mlab.options.backend = 'envisage'

f_strain = mlab.figure()
scene = mlab.get_engine().scenes[-1]
scene.name = 'strain'
strain_viz = Viz3DTensorField(vis3d=s.hist['strain'])
strain_viz.setup()

decorate_figure(f_strain, strain_viz, 200, [70, 20, 0])

mlab.show()



 
from traits.api import \
    HasTraits, List, Array, Property, cached_property, \
    Instance, Trait, Button, on_trait_change, Tuple, \
    Int, Float, String

from traits.api import Enum

from traitsui.api import \
    View, Item

from ibvpy.core.i_sdomain import \
    ISDomain
    
from ibvpy.core.sdomain import \
    SDomain

from numpy import \
    array, unique, min, max, mgrid, ogrid, c_, alltrue, repeat, ix_, \
    arange, ones, zeros, multiply, sort

from ibvpy.plugins.mayavi_util.pipelines import \
    MVPolyData, MVPointLabels
  
class MGridCellSpec(HasTraits):
    '''
    '''
    
    geo_type = Enum("Triangle", "Diamond")

    
    node_coords = List( [[-1,   -1,  0],
                         [ 1,   -1,  0],
                         [ 1,    1,  0],
                         [ 1,    1,  1],
                         [-1,    1, -1],
                         [-1/2., 1,  0],
                         [ 0.  , 1,  0],
                         [ 1/2., 1,  0]] )

    xnode_coords = List( [[-1,0,-1],
                          [ 1,0,-1],
                          [ 1,0, 1],
                          [-1,0, 1]] )
    #-------------------------------------------------------------------
    # Visualization-related specification
    #-------------------------------------------------------------------
    cell_lines = Array( int, value = [[0,1],[1,2],[2,0]])

    @on_trait_change('geo_type')
    def _reset_node_coords(self):
        if self.geo_type == "Triangle":
            # three points selected like triangle 
            self.node_coords =  [[-1,-1,-1],
                                 [ 1, 0, 0],
                                 [ 1, 1, 1],
                                 ]
            # Triangle: 
            self.cell_lines = [[0,1],[1,2],[2,0]]
            cell_faces = [[0,1,2]]
        else:
            #Diamond
            self.node_coords = [[0,   -1,   -1],   
                                [1,    0,   -1],   
                                [0,    1,   -1],   
                                [-1,   0,   -1],
        
                                [0,   -1,    1],
                                [1,    0,    1],
                                [0,    1,    1],
                                [-1,   0,    1],
                                
                                [1,   -1,    0],
                                [-1,  -1,    0],
                                
                                [1,    1,    0],
                                [-1,   1,    0],
                                ]
            # Diamand
            self.cell_lines =  [[0,1],[1,2],[2,3],[3,0], 
                           [4,5],[5,6],[6,7],[7,4], 
                           [0,8],[8,4],[4,9],[9,0], 
                           [2,10],[10,6],[6,11],[11,2],
                           [9,3],[3,11],[11,7],[7,9],
                           [8,1],[1,10],[10,5],[5,8] ]
            
            cell_faces = [[0,3,2,1],[4,5,6,7], [0,8,4,9], [2,10,6,11], [9,3,11,7], [8,1,10,5] ]


    traits_view = View(Item('geo_type'),
                       Item('node_coords',style='readonly'),
                       Item('cell_lines',style='readonly'),
                       resizable = True,
                       height = 0.5,
                       width = 0.5)

    _node_array = Property( Array('float_'), depends_on = 'node_coords' )
    @cached_property
    def _get__node_array(self):
        '''Get the node array as float_
        '''
        # check that the nodes are equidistant
        return array( self.node_coords, 'float_' )

    def get_n_dims(self):
        '''Get the number of dimension of the cell
        '''
        return self._node_array.shape[1]
    
    def get_cell_shape(self):
        '''Get the shape of the cell grid.
        '''
        npoints = self._node_array.shape[1]
        narray = self._node_array

        return array( [len( unique( narray[:,i] ) )
                       for i in range(npoints) ], dtype = int )

    def get_cell_slices(self):
        '''Get slices for the generation of the cell grid.
        '''
        ndims = self.get_n_dims()
        narray = self._node_array
        return tuple( [ slice( 
                              min( narray[:,i] ),
                              max( narray[:,i] ),
                              complex(0,len(unique(narray[:,i]))), 
                              ) 
                        for i in range(ndims) ] )

class MGridCell(SDomain):
    '''
    A single mgrid cell for geometrical representation of the domain.
    
    Based on the grid_cell_spec attribute, 
    the node distribution is determined.
    
    '''
    # Everything depends on the grid_cell_specification
    #
    grid_cell_spec = Instance( MGridCellSpec )
    def _grid_cell_spec_default(self):
        return MGridCellSpec()
    
    
    # Generated grid cell coordinates as they come from mgrid.
    # The dimensionality of the mgrid comes from the 
    # grid_cell_spec_attribute
    #
    grid_cell_coords = Property( depends_on = 'grid_cell_spec')
    @cached_property
    def _get_grid_cell_coords(self):
        grid_cell = mgrid[ self.grid_cell_spec.get_cell_slices() ]
        return c_[ tuple([ x.flatten() for x in grid_cell ]) ]

    # Node map lists the active nodes within the grid cell
    # in the specified order
    #
    node_map = Property( Array(int), depends_on = 'grid_cell_spec' )
    @cached_property
    def _get_node_map(self):
        n_map = []
        for node in self.grid_cell_spec._node_array:
            for idx, grid_cell_node in enumerate( self.grid_cell_coords ):
                if alltrue( node == grid_cell_node ):
                    n_map.append( idx )
                    continue
        return array(n_map,int)

    #-----------------------------------------------------------------
    # Visualization related methods
    #-----------------------------------------------------------------
    mvp_mgrid_ngeo_labels = Trait( MVPointLabels )
    def _mvp_mgrid_ngeo_labels_default(self):
        return MVPointLabels( name = 'Geo node numbers',
                                  points = self._get_points,
                                  scalars = self._get_node_distribution,
                                  color = (153,204,0))
        
     
    refresh_button = Button('Draw')
    @on_trait_change('refresh_button')
    def redraw(selfMGridCell):
        '''
        '''
        self.mvp_mgrid_ngeo_labels.redraw( 'label_scalars' )        

    def _get_points(self):
        #points = self.grid_cell_coords[ ix_(self.node_map) ]

        points = self.grid_cell_coords 
        print(points)
        shape = points.shape
        if shape[1] < 3:
            _points = zeros( (shape[0],3), dtype = float )
            _points[:,0:shape[1]] = points
            return _points
        else:
            return points

    def _get_node_distribution(self):
        #return arange(len(self.node_map))
        n_points = self.grid_cell_coords.shape[0]
        full_node_map = ones(n_points,dtype=float) * -1.
        full_node_map[ ix_(self.node_map) ] = arange(len(self.node_map))
        print(full_node_map)
        return full_node_map

    #------------------------------------------------------------------
    # UI - related methods
    #------------------------------------------------------------------
    traits_view = View(Item('grid_cell_spec'),
                       Item('refresh_button'),
                       Item('node_map'),
                       resizable = True,
                       height = 0.5,
                       width = 0.5)        

class MGridPntCell(MGridCell):
    '''
    '''
    
class MGridDofCell(MGridCell):
    '''
    '''


if __name__ == '__main__':
    
    mgc = MGridCell( grid_cell_spec = MGridCellSpec() )
    
    from ibvpy.plugins.ibvpy_app import IBVPyApp
    ibvpy_app = IBVPyApp( ibv_resource = mgc )
    ibvpy_app.main()




#-------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Sep 8, 2009 by: rch

import math

from ibvpy.tmodel.mats1D5.mats1D5_eval import MATS1D5Eval
from ibvpy.tmodel.mats_eval import IMATSEval
from ibvpy.mathkit.numpy.numpy_func import Heaviside
from traits.api import \
    Instance, Property, cached_property, List, \
    Callable, String, Int, HasTraits, Float
from traitsui.api import \
    View, Item

import numpy as np


class MATS1D5PressureSensitive(MATS1D5Eval):
    r'''Bond model for two phases interacting over an interface with zero thickness.
The frictional resistance is governed by the current level of pressure 
using the cohesion :math:`c` and frictional angle :math`\phi`. 

Basic formulas

The threshold level of frictional stress is prescribed by the Mohr-Coulomb rule:

.. math::
    \tau_\mathrm{fr}(\sigma) = \left( c + \sigma \; \tan{(\phi)} \right) 
     H\left( \sigma -  \frac{c}{\tan{(\phi)}}\right)
    :label: eq_MATS1D5PressureSensitive

The deformation variables include interface opening :math:`w` and sliding :math:`s`
related to the stress variables using the equations

.. math::
    \tau = G_s \cdot ( s - s^p )

.. math::
    \sigma = \left( G_w^{(+)} H(w) + G_w^{(-)} H(-w) \right) \cdot w

where 
 - :math:`G_s` represents elastic shear stiffness, 
 - :math:`G_w^{(+)}` is the adhesive stiffness for crack opening and 
 - :math:`G_w^{(-)}` is the penetration (or penalty) stiffness to be set large.

The elastic domain is given by the yield condition

.. math::
    f := \left| \tau \right| - \tau_\mathrm{fr}(\sigma) \leq 0

Thus, the Kuhn-Tucker condition must hold:

.. math::
    \dot{\gamma} \geq 0, \; f( \sigma, \tau ) \cdot \gamma \leq 0, \;
    \dot{\gamma} f(\sigma, \tau) = 0

where :math:`\gamma` represents the plastic multiplier.     
The flow rule for yielding slip is provided as

.. math::
    \dot{s}_p = \dot{\gamma} \; \mathrm{sign}(\tau)

and the consistency condition (staying on the yield surface upon platic loading)

.. math::
    \dot{\gamma} \dot{f}(\sigma, \tau) = 0

Discrete form

Using the midpoint rule 

Given the displacement increment 

.. math::
    s_{n+1} = s_{n} + \Delta s_{n}, \; 
    w_{n+1} = w_{n} + \Delta w_{n}, \;
    :label: eq_increment

the elastic trial stresses are calculated as

.. math::
    \tau_{n+1}^{\mathrm{trial}} = G_s ( s_{n+1} - s_n^p ), \;
    \sigma_{n+1} = G_w ( w_{n+1} )
    :label: eq_constitutive_law

and the frictional threshold for the current transverse stresses

.. math::
    \tau^\mathrm{fr}_{n+1}(\sigma_{n+1}) = 
    \left( c + \sigma_{n+1} \; \tan{(\phi)} \right) 
     H\left( \sigma_{n+1} -  \frac{c}{\tan{(\phi)}}\right)
    :label: eq_mohr_coulomb

The trial consistency condition is obtained as

.. math::
    f_{n+1}^\mathrm{trial} = \left| \tau_{n+1}^\mathrm{trial} \right|
    - \tau^\mathrm{fr}_{n+1}(\sigma_{n+1})
    :label: eq_discr_consistency

If :math:`f^\mathrm{trial}_{n+1} \leq 0` then step is elastic.

If :math:`f^\mathrm{trial}_{n+1} > 0 \Leftrightarrow f(\tau_{n+1}, \sigma_{n+1} ) = 0`
The objective is to identify :math:`s^{p}_{n+1}, \Delta \gamma` satisfying this condition
and the condition of consistency. To accomplish this task we first note that

.. math::
    \tau_{n+1}= G_s( s_{n+1} - s^p_{n+1} ) =G_s( s_{n+1} - s^p_{n}) - G_s( s^p_{n+1} - s^p_n )

Using Eq.:eq:`eq_midpoint` the shear stress can be related to trial state and the plastic 
(return mapping) multiplier :math:`\Delta \gamma` as 

.. math::
    \tau_{n+1} = \tau^{\mathrm{trial}}_{n+1} - G_s( s^p_{n+1} - s^p_n )
    = \tau^{\mathrm{trial}}_{n+1} - G_s \Delta \gamma \, \mathrm{sign}( \tau_{n+1} ).

The other two incremental equation deliver the plastic slip and consistency condition.

.. math::
    s^{p}_{n+1} = s^p_n + \Delta \gamma \, \mathrm{sign}( \tau_{n+1} )
    :label: eq_midpoint

.. math::
    f_{n+1} = \left| \tau_{n+1} \right| - \tau^{\mathrm{fr}}(\sigma_{n+1}) = 0
    :label: eq_discr_consistency2

It can be shown that the direction of mapping given by :math:`\mathrm{sign}(\tau_{n+1})`
is consistent with the trial state, i,e. :math:`\mathrm{sign}(\tau^{\mathrm{trial}}_{n+1})`
and

.. math::
    \left| \tau_{n+1} \right| + \Delta \gamma G_s = \left| \tau^{\mathrm{trial}}_{n+1} \right|

Finally, due to :math:`\Delta \gamma > 0` the discrete consistency condition 
(:eq:`eq_discr_consistency`) can be further expanded as

.. math::
    f_{n+1} = \left| \tau^{\mathrm{trial}}_{n+1} \right| - 
    G_s \Delta \gamma- \tau^{\mathrm{fr}}( \sigma_{n+1} )

Hence

.. math::
    f_{n+1} = 0 \implies \Delta \gamma = \frac{ f^{\mathrm{trial}}_{n+1}}{ G_s } > 0

and

.. math::
    \tau_{n+1} &= \tau^{\mathrm{trial}}_{n+1} - 
    \Delta \gamma G_s \, \mathrm{sign}( \tau_{n+1}^{\mathrm{trial}} ) \\
    s^p_{n+1} &= s^p_n + 
    \Delta \gamma \, \mathrm{sign}( \tau_{n+1}^{\mathrm{trial}} ) \\ 

    '''

    # implements(IMATSEval)

    G_s = Float(1.0, input=True, enter_set=False,
                label='cohesion')

    G_w_open = Float(1.0, input=True, enter_set=False,
                     label='cohesion')

    G_w_close = Float(1e+6, input=True, enter_set=False,
                      label='cohesion')

    c = Float(0.0, input=True, enter_set=False,
              label='cohesion')

    phi = Float(0.1, input=True, enter_set=False,
                label='friction angle')

    #-------------------------------------------------------------------------
    # Submodels constituting the interface behavior
    #-------------------------------------------------------------------------

    traits_view = View(Item('c@'),
                       Item('phi@'),
                       Item('G_open@'),
                       Item('G_slip@'),
                       Item('G_penalty@'),
                       resizable=True,
                       scrollable=True,
                       width=0.8,
                       height=0.9,
                       buttons=['OK', 'Cancel'])

    #-------------------------------------------------------------------------
    # Setup for computation within a supplied spatial context
    #-------------------------------------------------------------------------

    state_var_shapes = {'slip_p': ()}

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------

    def get_corr_pred(self, sctx, u_r, tn1, slip_p):
        '''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain
        '''

        s_n = u_r[..., 0]
        w = u_r[..., 1]
        s_p_n = slip_p

        f_separate = (Heaviside(w) * self.G_open + 
                      Heaviside(-w) * self.G_penalty) * w

        f_r = np.zeros_like(u_r)
        tau_trial = self.G_slip * (s_n - s_p_n)
        f_trial = abs(tau_trial) - (f_separate + math.tan(self.phi))

        f_I = np.where(f_trial > 0.0)
        f_r[..., 0] = tau_trial

        D_n1[0, 0] = E
        d_gamma = f_trial / (self.E + self.K_bar + self.H_bar)
        sig_n1[0] = sigma_trial - d_gamma * self.E * sign(xi_trial)
        D_n1[0, 0] = (self.E * (self.K_bar + self.H_bar)) / \
            (self.E + self.K_bar + self.H_bar)

        sig_n1 = zeros((1,), dtype='float_')
        D_n1 = zeros((1, 1), dtype='float_')

        D_mtx = zeros(
            (eps_app_eng.shape[0], eps_app_eng.shape[0]), dtype='float_')

        return sig_app_eng, D_mtx

    def get_sig1(self, sctx, eps_app_eng, *args, **kw):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return sig_eng[0:1]

    def get_sig2(self, sctx, eps_app_eng, *args, **kw):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return sig_eng[3:]

    def get_shear_flow(self, sctx, eps_app_eng, *args, **kw):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return sig_eng[1:2]

    def get_cohesive_stress(self, sctx, eps_app_eng, *args, **kw):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return sig_eng[2:3]

    rte_dict = Property

    def _get_rte_dict(self):

        rte_dict = {}
        ix_maps = [0, 1, 2, 3]
        for name, mats, ix_map, size, offset in \
            zip(self._mats_names, self._mats_list, ix_maps, self._state_sizes,
                self._state_offsets):
            for key, v_eval in list(mats.rte_dict.items()):

                __call_v_eval = RTE1D5Bond(v_eval=v_eval,
                                           name=name + '_' + key,
                                           size=size,
                                           offset=offset,
                                           ix_map=ix_map)

                rte_dict[name + '_' + key] = __call_v_eval

        # sigma is also achievable through phase1_sig_app and phase_2_sig_app
        extra_rte_dict = {'sig1': self.get_sig1,
                          'sig2': self.get_sig2,
                          'shear_flow': self.get_shear_flow,
                          'cohesive_stress': self.get_cohesive_stress,
                          }
        rte_dict.update(extra_rte_dict)
        return rte_dict
    #-------------------------------------------------------------------------
    # Methods required by the mats_explore tool
    #-------------------------------------------------------------------------

    def new_cntl_var(self):
        return zeros(4, 'float_')

    def new_resp_var(self):
        return zeros(4, 'float_')


class RTE1D5Bond(HasTraits):

    v_eval = Callable
    name = String
    size = Int
    offset = Int
    ix_map = Int

    def __call__(self, sctx, u, *args, **kw):
        u_x = array([u[self.ix_map]], dtype='float')
        # save the spatial context
        mats_state_array = sctx.mats_state_array
        sctx.mats_state_array = mats_state_array[self.offset: self.offset + self.size]
        result = self.v_eval(sctx, u_x, *args, **kw)
        # put the spatial context back
        sctx.mats_state_array = mats_state_array

        return result


def sp_derive():

    import sympy as sp

    vars = 'G_s, s_n, s_p_n, w_n, dw_n, ds_n, G_s, G_w, c, phi'

    syms = sp.symbols(vars)

    for var, sym in zip(vars.split(','), syms):
        globals()[var.strip()] = sym

    s_n1 = s_n + ds_n
    w_n1 = w_n + dw_n

    tau_trial = G_s * (s_n1 - s_p_n)

    print('diff', sp.diff(tau_trial, ds_n))

    print(tau_trial)

    sig_n1 = G_w * w_n1

    print(sig_n1)

    tau_fr = (c + sig_n1 * sp.tan(phi)) * \
        sp.Heaviside(sig_n1 - c / sp.tan(phi))

    print(tau_fr)

    d_tau_fr = sp.diff(tau_fr, dw_n)

    print(d_tau_fr)

    f_trial = sp.abs(tau_trial) - tau_fr

    print(f_trial)

    d_gamma = f_trial / G_s

    print('d_gamma')
    sp.pretty_print(d_gamma)

    print('d_gamma_s')
    sp.pretty_print(sp.diff(d_gamma, ds_n))

    print('tau_n1')
    tau_n1 = sp.simplify(tau_trial - d_gamma * G_s * sp.sign(tau_trial))
    sp.pretty_print(tau_n1)

    print('dtau_n1_w')
    dtau_n1_w = sp.diff(tau_n1, dw_n)
    sp.pretty_print(dtau_n1_w)

    print('dtau_n1_s')
    dtau_n1_s = sp.diff(d_gamma * sp.sign(tau_trial), ds_n)
    print(dtau_n1_s)

    s_p_n1 = s_p_n + d_gamma * sp.sign(tau_trial)

    print(s_p_n1)


if __name__ == '__main__':

    sp_derive()




import time

from simulator.api import \
    TStep, Simulator
from simulator.tloop_implicit import TLoopImplicit
from traits.api import \
    Float
import traits.has_traits

import numpy as np

from .interaction_scripts import run_rerun_test


traits.has_traits.CHECK_INTERFACES = 2


class DemoQuadFNModel(TStep):
    '''Model implementing both the value of the residual of the governing
    equations and of their derivatives to be plugged into the implicit
    time stepping algorithm.
    '''
    tloop_type = TLoopImplicit

    R_0 = Float(1.0, auto_set=False, enter_set=True)
    '''Target value of a function (load).
    '''

    def get_corr_pred(self, U_k, t_n1):
        '''Return the value and the derivative of a function
        '''
        R = U_k * U_k - (self.R_0 * t_n1)
        dR = 2 * U_k
        # handle the case of zero derivative - return small number
        dR = max(np.fabs(dR), 1e-3)
        # sleep a bit to make the calculation slower - sleep for 100 msecs
        # to demonstrate the pause-resume, stop-restart functionality
        time.sleep(0.01)
        return R, dR


# Construct a Simulator
m = DemoQuadFNModel()
s = m.sim
run_rerun_test(s)


'''

This script is used to demonstrate the states of a model.

@author: rch
'''

import time

from simulator.api import \
    Simulator, TLoop, TStep
import traits.has_traits

from .interaction_scripts import run_rerun_test


traits.has_traits.CHECK_INTERFACES = 2


class DemoExplicitTLoop(TLoop):
    '''Demonstration loop running explicitly evaluating
    a function and saving the results in the history.
    This kind of time loop is used e.g. for the explicit
    calculation of stress strain curve with all components
    prescribed.
    '''

    def eval(self):
        t_n = self.tline.val
        t_max = self.tline.max
        dt = self.tline.step
        while t_n < t_max:
            print('\ttime %g' % t_n)
            if self.user_wants_abort:
                break
            time.sleep(1)
            t_n += dt
            self.hist.record_timestep(t_n, None, None)
            self.tline.val = t_n
        return


class DemoStatesModel(TStep):
    tloop_type = DemoExplicitTLoop


# Construct a Simulator
m = DemoStatesModel()
s = m.sim
run_rerun_test(s)




from math import pi as Pi, cos, sin, exp, sqrt as scalar_sqrt

from ibvpy.api import RTrace, RTDofGraph, RTraceArraySnapshot
from ibvpy.core.tstepper import \
     TStepper as TS
from ibvpy.tmodel.mats_eval import IMATSEval, MATSEval
from mathkit.mfn import MFnLineArray
from mathkit.mfn.mfn_line.mfn_matplotlib_editor import MFnMatplotlibEditor
from mathkit.mfn.mfn_line.mfn_plot_adapter import MFnPlotAdapter
from numpy import \
     array, ones, zeros, outer, inner, transpose, dot, frompyfunc, \
     fabs, sqrt, linspace, vdot, identity, tensordot, \
     sin as nsin, meshgrid, float_, ix_, \
     vstack, hstack, sqrt as arr_sqrt
from scipy.linalg import eig, inv
from traits.api import \
     Array, Bool, Callable, Enum, Float, HasTraits, \
     Instance, Int, Trait, Range, HasTraits, on_trait_change, Event, \
     Dict, Property, cached_property, Delegate
from traitsui.api import \
     Item, View, HSplit, VSplit, VGroup, Group, Spring

# from dacwt import DAC
mpl_matplotlib_editor = MFnMatplotlibEditor()

#---------------------------------------------------------------------------
# Material time-step-evaluator for Scalar-Damage-Model
#---------------------------------------------------------------------------


class MATS1D5Bond(MATSEval):
    '''
    Scalar Damage Model.
    '''
    # implements(IMATSEval)

    Ef = Float(1.,  # 34e+3,
                 label="E",
                 desc="Young's Modulus Fiber",
                 auto_set=False, enter_set=True)
    Af = Float(1.,  # 34e+3,
                 label="A",
                 desc="Cross Section Fiber",
                 auto_set=False, enter_set=True)
    Em = Float(1.,  # 34e+3,
                 label="E",
                 desc="Young's Modulus Matrix",
                 auto_set=False, enter_set=True)
    Am = Float(1.,  # 34e+3,
                 label="A",
                 desc="Cross Section Matrix",
                 auto_set=False, enter_set=True)
    G = Float(1.,  # 34e+3,
                 label="G",
                 desc="Shear Stiffness",
                 auto_set=False, enter_set=True)
    bond_fn = Trait(MFnLineArray(ydata=[0, 1]),
                     label="Bond",
                     desc="Bond Function",
                     auto_set=False)

    traits_view = View(Item('Ef'),
                        Item('Af'),
                        Item('Em'),
                        Item('Am'),
                        Item('G'),
                        Item('bond_fn', editor=mpl_matplotlib_editor),
                        resizable=True,
                        scrollable=True,
                        height=0.5,
                        width=0.5,
                        buttons=['OK', 'Cancel']
                        )

    def _bond_fn_default(self):
        return MFnLineArray(xdata=[0., 1.], ydata=[0., 1.])

    # This event can be used by the clients to trigger an action upon
    # the completed reconfiguration of the material model
    #
    changed = Event

    #---------------------------------------------------------------------------------------------
    # View specification
    #---------------------------------------------------------------------------------------------

    #-----------------------------------------------------------------------------------------------
    # Private initialization methods
    #-----------------------------------------------------------------------------------------------

    #-----------------------------------------------------------------------------------------------
    # Setup for computation within a supplied spatial context
    #-----------------------------------------------------------------------------------------------

    def setup(self, sctx):
        '''
        Intialize state variables.
        '''

    def new_cntl_var(self):
        return zeros(4, float_)  # TODO: adapt for 4-5..

    def new_resp_var(self):
        return zeros(4, float_)  # TODO: adapt for 4-5..

    #-----------------------------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-----------------------------------------------------------------------------------------------

    def get_corr_pred(self, sctx, eps_app_eng, d_eps, tn, tn1):
        '''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain
        '''
        if eps_app_eng.shape[0] == 4:
            D_mtx = zeros((4, 4))
            sigma = zeros(4)
        elif eps_app_eng.shape[0] == 5:
            D_mtx = zeros((5, 5))
            sigma = zeros(5)
            D_mtx[2, 2] = self.bond_fn.get_diff(eps_app_eng[2])
            sigma[2] = self.bond_fn.get_value(eps_app_eng[2])  # tau_m
        else:
            print("MATS1D5Bond: Unsupported number of strain components")

        D_mtx[0, 0] = self.bond_fn.get_diff(eps_app_eng[0])
        D_mtx[1, 1] = self.bond_fn.get_diff(eps_app_eng[1])

        D_mtx[-2, -2] = self.Ef * self.Af
        D_mtx[-1, -1] = self.Em * self.Am

        sigma[0] = self.bond_fn.get_value(eps_app_eng[0])  # tau_l
        sigma[1] = self.bond_fn.get_value(eps_app_eng[1])  # tau_r

        sigma[-2] = self.Ef * self.Af * eps_app_eng[-2]  # sig_f
        sigma[-1] = self.Em * self.Am * eps_app_eng[-1]  # sig_m

        # You print the stress you just computed and the value of the apparent E

        return  sigma, D_mtx

    #---------------------------------------------------------------------------------------------
    # Subsidiary methods realizing configurable features
    #---------------------------------------------------------------------------------------------

    def get_sig_app(self, sctx, eps_app_eng, *args, **kw):
        # @TODO
        # the stress calculation is performed twice - it might be
        # cached.
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        r_pnt = sctx.loc
        s_tensor = zeros((3, 3))
        if r_pnt[1] >= 0.:
            s_tensor[0, 0] = sig_eng[-1]
        else:
            s_tensor[0, 0] = sig_eng[-2]
        return s_tensor

    def get_shear(self, sctx, eps_app_eng):
        # @TODO
        # the stress calculation is performed twice - it might be
        # cached.
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        r_pnt = sctx.loc
        if eps_app_eng.shape[0] == 4:
            if r_pnt[0] >= 0.:
                sig_pos = [sig_eng[1]]
            else:
                sig_pos = [sig_eng[0]]
        elif eps_app_eng.shape[0] == 5:
            if r_pnt[0] >= 0.5:
                sig_pos = [sig_eng[1]]
            elif r_pnt[0] <= -0.5:
                sig_pos = [sig_eng[0]]
            else:
                sig_pos = [sig_eng[2]]
        return sig_pos

    def get_eps_app(self, sctx, eps_app_eng):
        r_pnt = sctx.loc
        if r_pnt[1] >= 0.:
            eps_pos = [eps_app_eng[-1]]
        else:
            eps_pos = [eps_app_eng[-2]]
        return eps_pos

    def get_slip(self, sctx, eps_app_eng):
        r_pnt = sctx.loc
        if eps_app_eng.shape[0] == 4:
            if r_pnt[0] >= 0.:
                eps_pos = [eps_app_eng[1]]
            else:
                eps_pos = [eps_app_eng[0]]
        if eps_app_eng.shape[0] == 5:
            if r_pnt[0] >= 0.5:
                eps_pos = [eps_app_eng[1]]
            elif r_pnt[0] <= -0.5:
                eps_pos = [eps_app_eng[0]]
            else:
                eps_pos = [eps_app_eng[2]]
        return eps_pos

    #---------------------------------------------------------------------------------------------
    # Response trace evaluators
    #---------------------------------------------------------------------------------------------

    # Declare and fill-in the rte_dict - it is used by the clients to
    # assemble all the available time-steppers.
    #
    rte_dict = Trait(Dict)

    def _rte_dict_default(self):
        return {'sig_app_t1d'      : self.get_sig_app,
                'eps_app_t1d'      : self.get_eps_app,
                'shear_s'        : self.get_shear,
                'slip_s'         : self.get_slip,
                'msig_pos'     : self.get_msig_pos,
                'msig_pm'      : self.get_msig_pm}


if __name__ == '__main__':
     mats_bond = MATS1D5Bond()
     mats_bond.configure_traits()




from math import pi as Pi, cos, sin, exp, sqrt as scalar_sqrt

from ibvpy.api import RTrace, RTDofGraph, RTraceArraySnapshot
from ibvpy.core.tstepper import \
     TStepper as TS
from ibvpy.tmodel.mats_eval import IMATSEval, MATSEval
from ibvpy.mathkit.mfn import MFnLineArray
from ibvpy.mathkit.mfn.mfn_line.mfn_matplotlib_editor import MFnMatplotlibEditor
from ibvpy.mathkit.mfn.mfn_line.mfn_plot_adapter import MFnPlotAdapter
from numpy import \
     array, ones, zeros, outer, inner, transpose, dot, frompyfunc, \
     fabs, sqrt, linspace, vdot, identity, tensordot, \
     sin as nsin, meshgrid, float_, ix_, \
     vstack, hstack, sqrt as arr_sqrt
from scipy.linalg import eig, inv
from traits.api import \
     Array, Bool, Callable, Enum, Float, HasTraits, \
     Instance, Int, Trait, Range, HasTraits, on_trait_change, Event, \
     Dict, Property, cached_property, Delegate
from traitsui.api import \
     Item, View, HSplit, VSplit, VGroup, Group, Spring

# from dacwt import DAC
mpl_matplotlib_editor = MFnMatplotlibEditor(
                    adapter=MFnPlotAdapter(label_x='slip',
                                              label_y='shear stress'))

#---------------------------------------------------------------------------
# Material time-step-evaluator for Scalar-Damage-Model
#---------------------------------------------------------------------------


class MATS1D5Bond(MATSEval):
    '''
    Scalar Damage Model.
    '''
    # implements(IMATSEval)

    Ef = Float(1.,  # 34e+3,
                 label="E",
                 desc="Young's Modulus Fiber",
                 auto_set=False, enter_set=True)
    Af = Float(1.,  # 34e+3,
                 label="A",
                 desc="Cross Section Fiber",
                 auto_set=False, enter_set=True)
    Em = Float(1.,  # 34e+3,
                 label="E",
                 desc="Young's Modulus Matrix",
                 auto_set=False, enter_set=True)
    Am = Float(1.,  # 34e+3,
                 label="A",
                 desc="Cross Section Matrix",
                 auto_set=False, enter_set=True)

    s_cr = Float(1.,  # 34e+3,
                 label="s_cr",
                 desc="Critical slip",
                 auto_set=False, enter_set=True)

    perimeter = Property(depends_on='A')

    def _get_perimeter(self):
        return sqrt(4 * self.Af * Pi)

    slip_middle = Bool
    #---------------------------------------------------------------------------------
    # Bond stress (unit area)
    #---------------------------------------------------------------------------------
    tau_max = Float(1.,  # 34e+3,
                 label="tau_max",
                 desc="maximum shear stress",
                 auto_set=False, enter_set=True)
    tau_fr = Float(1.,  # 34e+3,
                 label="tau_fr",
                 desc="Frictional shear stress",
                 auto_set=False, enter_set=True)

    #---------------------------------------------------------------------------------
    # Bond flow (per unit length) - under the assumption of circular cross section
    #---------------------------------------------------------------------------------
    T_max = Property(depends_on='tau_max, Af')

    @cached_property
    def _get_T_max(self):
        return self.tau_max * self.perimeter

    T_fr = Property(depends_on='tau_fr, Af')

    @cached_property
    def _get_T_fr(self):
        return self.tau_fr * self.perimeter

    #---------------------------------------------------------------------------------
    # Bond flow (per unit length) - under the assumption of circular cross section
    #---------------------------------------------------------------------------------
    bond_fn = Property(depends_on='s_cr, tau_max, tau_fr, Af')

    @cached_property
    def _get_bond_fn(self):
        return MFnLineArray(ydata=[ 0., self.T_max, self.T_fr, self.T_fr ],
                             xdata=[ 0., self.s_cr, self.s_cr, self.s_cr * 10])

    traits_view = View(Item('Ef'),
                        Item('Af'),
                        Item('Em'),
                        Item('Am'),
                        Item('tau_fr'),
                        Item('s_cr'),
                        Item('tau_max'),
                        Item('bond_fn', editor=mpl_matplotlib_editor),
                        resizable=True,
                        scrollable=True,
                        height=0.5,
                        width=0.5,
                        buttons=['OK', 'Cancel']
                        )

    def _bond_fn_default(self):
        return MFnLineArray(xdata=[0., 1.], ydata=[0., 1.])

   # This event can be used by the clients to trigger an action upon
    # the completed reconfiguration of the material model
    #
    changed = Event

    #---------------------------------------------------------------------------------------------
    # View specification
    #---------------------------------------------------------------------------------------------

    #-----------------------------------------------------------------------------------------------
    # Private initialization methods
    #-----------------------------------------------------------------------------------------------

    #-----------------------------------------------------------------------------------------------
    # Setup for computation within a supplied spatial context
    #-----------------------------------------------------------------------------------------------

    def setup(self, sctx):
        '''
        Intialize state variables.
        '''
        sctx.mats_state_array = zeros(sctx.slip_comp, float_)

    def get_state_array_size(self):
        '''
        Give back the nuber of floats to be saved
        @param sctx:spatial context
        '''
        return 2

    def new_cntl_var(self):
        return zeros(4, float_)  # TODO: adapt for 4-5..

    def new_resp_var(self):
        return zeros(4, float_)  # TODO: adapt for 4-5..

    #-----------------------------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-----------------------------------------------------------------------------------------------

    def get_corr_pred(self, sctx, eps_app_eng, d_eps, tn, tn1):
        '''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain

        @todo: [Jakub]
        - Describe the state variable vector - what is at what position
        - Describe the components of the strain tensor
        - In all algebraic expressions leave spaces between operators and operands
          a = b and not a=b
        - Comment every relational condition
          if dbond[1] > 0:
        '''
        if eps_app_eng.shape[0] == 5: self.slip_middle = True

        if sctx.update_state_on:
#            print 'in us'
#            print 'eps_app_eng ',eps_app_eng
#            print 'd_eps ',d_eps
            eps_n = eps_app_eng - d_eps
            sctx.mats_state_array[:] = self._get_state_variables(sctx, eps_n)

        dbond = self._get_state_variables(sctx, eps_app_eng)

        if self.slip_middle:
            D_mtx = zeros((5, 5))
            sigma = zeros(5)
            if dbond[2] > 0.:
                D_mtx[2, 2] = 0.
                sigma[2] = self.T_fr
            else:
                D_mtx[2, 2] = self.bond_fn.get_diff(eps_app_eng[2])
                sigma[2] = self.bond_fn.get_value(eps_app_eng[2])  # tau_m
        else:
            D_mtx = zeros((4, 4))
            sigma = zeros(4)

        if dbond[0] > 0.:  # bond_l  
            D_mtx[0, 0] = 0.
            sigma[0] = self.T_fr
        else:
            D_mtx[0, 0] = self.bond_fn.get_diff(eps_app_eng[0])
            sigma[0] = self.bond_fn.get_value(eps_app_eng[0])  # tau_l

        if dbond[1] > 0.:  # bond_r  
            D_mtx[1, 1] = 0.
            sigma[1] = self.T_fr
        else:
            D_mtx[1, 1] = self.bond_fn.get_diff(eps_app_eng[1])
            sigma[1] = self.bond_fn.get_value(eps_app_eng[1])  # tau_r

        D_mtx[-2, -2] = self.Ef * self.Af
        D_mtx[-1, -1] = self.Em * self.Am

        sigma[-2] = self.Ef * self.Af * eps_app_eng[-2]  # sig_f
        sigma[-1] = self.Em * self.Am * eps_app_eng[-1]  # sig_m

        # You print the stress you just computed and the value of the apparent E

        return  sigma, D_mtx

    def _get_state_variables(self, sctx, eps_app_eng):
        dbond_l, dbond_r = sctx.mats_state_array[:2]

        if eps_app_eng[0] > self.s_cr:
            dbond_l = 1.
        if eps_app_eng[1] > self.s_cr:
            dbond_r = 1.

        if self.slip_middle:
            dbond_m = sctx.mats_state_array[-1]
            if eps_app_eng[2] > self.s_cr:
                dbond_m = 1.
            return dbond_l, dbond_r, dbond_m
        else:
            return dbond_l, dbond_r

    #---------------------------------------------------------------------------------------------
    # Subsidiary methods realizing configurable features
    #---------------------------------------------------------------------------------------------

    def get_sig_app(self, sctx, eps_app_eng, *args, **kw):
        # @TODO
        # the stress calculation is performed twice - it might be
        # cached.
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        r_pnt = sctx.loc
        s_tensor = zeros((3, 3))
        if r_pnt[1] >= 0.:
            s_tensor[0, 0] = sig_eng[-1]
        else:
            s_tensor[0, 0] = sig_eng[-2]
        return s_tensor

    def get_shear(self, sctx, eps_app_eng):
        # @TODO
        # the stress calculation is performed twice - it might be
        # cached.
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        r_pnt = sctx.loc
        if eps_app_eng.shape[0] == 4:
            if r_pnt[0] >= 0.:
                sig_pos = [sig_eng[1]]
            else:
                sig_pos = [sig_eng[0]]
        elif eps_app_eng.shape[0] == 5:
            if r_pnt[0] >= 0.5:
                sig_pos = [sig_eng[1]]
            elif r_pnt[0] <= -0.5:
                sig_pos = [sig_eng[0]]
            else:
                sig_pos = [sig_eng[2]]
        return sig_pos

    def get_eps_app(self, sctx, eps_app_eng):
        r_pnt = sctx.loc
        if r_pnt[1] >= 0.:
            eps_pos = [eps_app_eng[-1]]
        else:
            eps_pos = [eps_app_eng[-2]]
        return eps_pos

    def get_slip(self, sctx, eps_app_eng):
        r_pnt = sctx.loc
        if eps_app_eng.shape[0] == 4:
            if r_pnt[0] >= 0.:
                eps_pos = [eps_app_eng[1]]
            else:
                eps_pos = [eps_app_eng[0]]
        if eps_app_eng.shape[0] == 5:
            if r_pnt[0] >= 0.5:
                eps_pos = [eps_app_eng[1]]
            elif r_pnt[0] <= -0.5:
                eps_pos = [eps_app_eng[0]]
            else:
                eps_pos = [eps_app_eng[2]]
        return eps_pos

    def get_debonding(self, sctx, eps_app_eng):
        r_pnt = sctx.loc
        if self.slip_middle:
            if r_pnt[0] >= -0.25 and r_pnt[1] <= 0.25:
                dbond = sctx.mats_state_array[2]
            elif r_pnt[0] > 0.25:
                dbond = sctx.mats_state_array[1]
            else:
                dbond = sctx.mats_state_array[0]
        else:
            if r_pnt[0] >= 0.:
                dbond = sctx.mats_state_array[1]
            else:
                dbond = sctx.mats_state_array[0]
        return array([dbond], dtype=float_)

    #---------------------------------------------------------------------------------------------
    # Response trace evaluators
    #---------------------------------------------------------------------------------------------

    # Declare and fill-in the rte_dict - it is used by the clients to
    # assemble all the available time-steppers.
    #
    rte_dict = Trait(Dict)

    def _rte_dict_default(self):
        return {'sig_app_t1d'      : self.get_sig_app,
                'eps_app_t1d'      : self.get_eps_app,
                'shear_s'        : self.get_shear,
                'slip_s'         : self.get_slip,
                'debonding'      : self.get_debonding,
                'msig_pos'     : self.get_msig_pos,
                'msig_pm'      : self.get_msig_pm}


if __name__ == '__main__':
     mats_bond = MATS1D5Bond()
     mats_bond.configure_traits()



import time

from ibvpy.bcond import BCSlice
from ibvpy.fets import FETS2D4Q
from ibvpy.tmodel.mats3D.mats3D_elastic import MATS3DElastic
from ibvpy.tmodel.mats3D.mats3D_plastic.vmats3D_desmorat import \
    MATS3DDesmorat
from ibvpy.tmodel.viz3d_scalar_field import \
    Vis3DStateField, Viz3DScalarField
from ibvpy.tmodel.viz3d_tensor_field import \
    Vis3DTensorField, Viz3DTensorField
from mayavi import mlab
from simulator.api import \
    TStepBC
from simulator.xdomain.xdomain_fe_grid_axisym import XDomainFEGridAxiSym

import numpy as np

from .mlab_decorators import decorate_figure

inner_radius = 1
outer_radius = 2

xdomain = XDomainFEGridAxiSym(coord_min=(0, inner_radius),
                              coord_max=(1, outer_radius),
                              shape=(10, 10),
                              integ_factor=2 * np.pi,
                              fets=FETS2D4Q())

print(xdomain.B1_Eimabc.shape)
print(xdomain.B0_Eimabc.shape)

m = MATS3DDesmorat()
m = MATS3DElastic(E=1, nu=0.3)

left_y = BCSlice(slice=xdomain.mesh[:, 0, :, 0],
                 var='u', dims=[1], value=0)
left_x = BCSlice(slice=xdomain.mesh[0, :, 0, :],
                 var='u', dims=[0], value=-1.0)
right_x = BCSlice(slice=xdomain.mesh[-1, :, -1, :],
                  var='u', dims=[0], value=0.0)

m = TStepBC(
    domains=[(xdomain, m)],
    bc=[left_x, right_x, left_y],
)

m.hist.vis_record = {
    'strain': Vis3DTensorField(var='eps_ab'),
    'stress': Vis3DTensorField(var='sig_ab'),
    #        'damage': Vis3DStateField(var='omega_a'),
    #        'kinematic hardening': Vis3DStateField(var='z_a')
}

s = m.sim
s.tloop.k_max = 1000
s.tline.step = 0.1
s.tloop.verbose = True
s.run()

F_to = m.hist.F_t
U_to = m.hist.U_t
F_right_t = np.sum(F_to[:, right_x.dofs], axis=-1)
F_left_t = np.sum(F_to[:, left_x.dofs], axis=-1)
U_left_t = np.average(U_to[:, left_x.dofs], axis=-1)
print('left')
print('right')
print('U_left')
print(U_left_t)

mlab.options.backend = 'envisage'

f_strain = mlab.figure()
scene = mlab.get_engine().scenes[-1]
scene.name = 'strain'
strain_viz = Viz3DTensorField(vis3d=m.hist['strain'])
strain_viz.setup()

decorate_figure(f_strain, strain_viz, 200, [70, 20, 0])

f_stress = mlab.figure()
scene = mlab.get_engine().scenes[-1]
scene.name = 'stress'
stress_viz = Viz3DTensorField(vis3d=m.hist['stress'])
stress_viz.setup()

decorate_figure(f_stress, stress_viz, 200, [70, 20, 0])

mlab.show()




from ibvpy.sim.tline import TLine

from .bmcs_tree_view_handler import \
    BMCSTreeViewHandler, menu_save, \
    menu_open, plot_self
from .bmcs_viz_sheet import PlotPerspective
from .bmcs_window import \
    BMCSWindow, tree_node, tree_editor, \
    leaf_node



'''
Created on 25.03.2019

@author: fseemab
'''
import time

from ibvpy.bcond import BCSlice
from ibvpy.fets import FETS2D4Q
from ibvpy.fets.fets1D5 import FETS1D52ULRH
from ibvpy.tmodel.mats1D5.vmats1D5_d import \
    MATS1D5D
from ibvpy.tmodel.mats1D5.vmats1D5_dp import \
    MATS1D5DP
from ibvpy.tmodel.mats1D5.vmats1D5_dp_cum_press import \
    MATS1D5DPCumPress
from ibvpy.tmodel.mats3D.mats3D_elastic.vmats3D_elastic import \
    MATS3DElastic
from ibvpy.tmodel.mats3D.mats3D_plastic.vmats3D_desmorat import \
    MATS3DDesmorat
from ibvpy.tmodel.viz2d_field import \
    Vis2DField, Viz2DField
from ibvpy.tmodel.viz3d_scalar_field import \
    Vis3DStateField, Viz3DScalarField
from ibvpy.tmodel.viz3d_tensor_field import \
    Vis3DTensorField, Viz3DTensorField
from simulator.api import \
    Simulator
from simulator.xdomain.xdomain_fe_grid_axisym import XDomainFEGridAxiSym
from simulator.xdomain.xdomain_interface import XDomainFEInterface
from view.window import BMCSWindow

import numpy as np

from .viz2d_fw import Viz2DFW, Vis2DFW


#from .mlab_decorators import decorate_figure
ds = 14
n_e_ds = 12
phi = 2.5
L_b = phi * ds
n_b = phi * n_e_ds
L_e = ds / n_e_ds
n_ex = 4
L_e_ex = n_ex * L_e
n_y_concrete = 8
n_y_steel = 1
L_x = 2 * L_e_ex + L_b
n_e = 2 * n_ex + n_b
R_steel = ds / 2
R_concrete = 7 * ds
xd_steel_1 = XDomainFEGridAxiSym(coord_min=(0, 0),
                                 coord_max=(L_x, R_steel),
                                 shape=(n_e, n_y_steel),
                                 integ_factor=2 * np.pi,
                                 fets=FETS2D4Q())
print(np.max(xd_steel_1.x_Eia[..., 0]))
xd_concrete_2 = XDomainFEGridAxiSym(coord_min=(0, R_steel),
                                    coord_max=(L_x, R_concrete),
                                    shape=(n_e, n_y_concrete),
                                    integ_factor=2 * np.pi,
                                    fets=FETS2D4Q())

# m_steel = MATS3DDesmorat(E_1=210000, nu=0.3, tau_bar=2000.0)
m_steel = MATS3DElastic(E=210000, nu=0.2)
# m_concrete = MATS3DDesmorat(tau_bar=2.0)
m_concrete = MATS3DElastic(E=28000, nu=0.3)

xd12 = XDomainFEInterface(
    I=xd_steel_1.mesh.I[n_ex:-n_ex, -1],
    J=xd_concrete_2.mesh.I[n_ex:-n_ex, 0],
    fets=FETS1D52ULRH(),
    integ_factor=np.pi * ds
)

u_max = 0.3
#f_max = 4000.0
right_x_s = BCSlice(slice=xd_steel_1.mesh[-1, :, -1, :],
                    var='u', dims=[0], value=u_max)
right_x_c = BCSlice(slice=xd_concrete_2.mesh[0, 1:, 0, :],
                    var='u', dims=[0], value=0)
left_x_s = BCSlice(slice=xd_steel_1.mesh[0, :, 0, :],
                   var='f', dims=[0], value=0)

bc1 = [right_x_c, right_x_s]

tau_bar = 2.0
E_T = 1000
s_0 = tau_bar / E_T
print('s_0', s_0)
m_interface = MATS1D5DPCumPress(  # E_T=E_T, E_N=1000000, omega_fn_type='jirasek',
    algorithmic=True)
#m_interface.omega_fn.trait_set(s_0=s_0, s_f=1000 * s_0)

s = Simulator(
    domains=[(xd_steel_1, m_steel),
             (xd_concrete_2, m_concrete),
             (xd12, m_interface),
             ],
    bc=bc1,  # + bc2,
    record={
        'Pw': Vis2DFW(bc_right=right_x_s, bc_left=left_x_s),
        'slip': Vis2DField(var='slip'),
        'strain': Vis3DTensorField(var='eps_ab'),
        'stress': Vis3DTensorField(var='sig_ab'),
        #       'damage': Vis3DStateField(var='omega_a'),
        #        'kinematic hardening': Vis3DStateField(var='z_a')
    }
)

xd12.hidden = True

s.tloop.k_max = 1000
s.tline.step = 0.005
s.tloop.verbose = True
s.tstep.fe_domain.serialized_subdomains
fw = Viz2DFW(name='Pw', vis2d=s.hist['Pw'])
fslip = Viz2DField(name='slip', vis2d=s.hist['slip'])

w = BMCSWindow(sim=s)
w.viz_sheet.viz2d_list.append(fw)
w.viz_sheet.viz2d_list.append(fslip)
#w.offline = True
# w.configure_traits()

# mlab.options.backend = 'envisage'
#
# f_strain = mlab.figure()
# scene = mlab.get_engine().scenes[-1]
# scene.name = 'stress'
strain_viz = Viz3DTensorField(vis3d=s.hist['strain'])
# strain_viz.setup()
#strain_viz.warp_vector.filter.scale_factor = 100.0
# strain_viz.plot(s.tstep.t_n)
w.viz_sheet.add_viz3d(strain_viz)

# f_stress = mlab.figure()
# scene = mlab.get_engine().scenes[-1]
# scene.name = 'stress'
stress_viz = Viz3DTensorField(vis3d=s.hist['stress'])
# stress_viz.setup()
#stress_viz.warp_vector.filter.scale_factor = 100.0
# stress_viz.plot(s.tstep.t_n)
w.viz_sheet.add_viz3d(stress_viz)

# f_damage = mlab.figure()
# scene = mlab.get_engine().scenes[-1]
# scene.name = 'damage'
#damage_viz = Viz3DScalarField(vis3d=s.hist['damage'])
# damage_viz.setup()
#damage_viz.warp_vector.filter.scale_factor = 100.0
# damage_viz.plot(s.tstep.t_n)
# w.viz_sheet.add_viz3d(damage_viz)

w.run()
time.sleep(3)
# decorate_figure(f_strain, strain_viz)
# decorate_figure(f_stress, stress_viz)
# decorate_figure(f_damage, damage_viz)
# mlab.show()
w.configure_traits()



'''
Created on Apr 3, 2019

@author: rch
'''
from ibvpy.api import IBCond
from traits.api import \
    Bool, \
    Button, Str
from traits.api import \
    Tuple, HasStrictTraits, Array, WeakRef
from traitsui.api import \
    View, Item
from traitsui.ui_editors.array_view_editor import ArrayViewEditor
from view.plot2d import Viz2D, Vis2D

import numpy as np


class DataSheet(HasStrictTraits):

    data = Array(np.float_)

    view = View(
        Item('data',
             show_label=False,
             resizable=True,
             editor=ArrayViewEditor(titles=['x', 'y', 'z'],
                                    format='%.4f',
                                    # Font fails with wx in OSX;
                                    #   see traitsui issue #13:
                                    # font   = 'Arial 8'
                                    )
             ),
        width=0.5,
        height=0.6
    )


class Vis2DFW(Vis2D):

    Pw = Tuple()
    bc_right = Str
    bc_left = Str

    def _Pw_default(self):
        return ([0], [0], [0], [0])

    def update(self):
        sim = self.sim
        bc_right = sim.trait_get(self.bc_right)[self.bc_right]
        bc_left = sim.trait_get(self.bc_left)[self.bc_left]
        dofs_right = np.unique(bc_right.dofs)
        dofs_left = np.unique(bc_left.dofs)
        U_ti = sim.hist.U_t
        F_ti = sim.hist.F_t
        P = np.sum(F_ti[:, dofs_right], axis=1)
        P0 = np.sum(F_ti[:, dofs_left], axis=1)
        w = np.average(U_ti[:, dofs_right], axis=1)
        w0 = np.average(U_ti[:, dofs_left], axis=1)
        self.Pw = P, P0, w, w0


class Viz2DFW(Viz2D):
    '''Plot adaptor for the pull-out simulator.
    '''
    label = 'F-W'

    show_legend = Bool(True, auto_set=False, enter_set=True)

    def plot(self, ax, vot, *args, **kw):
        sim = self.vis2d.sim
        P_t, P0_t, w_t, w0_t = sim.hist['Pw'].Pw
        ymin, ymax = np.min(P_t), np.max(P_t)
        L_y = ymax - ymin
        ymax += 0.05 * L_y
        ymin -= 0.05 * L_y
        xmin, xmax = np.min(w_t), np.max(w_t)
        L_x = xmax - xmin
        xmax += 0.03 * L_x
        xmin -= 0.03 * L_x
        ax.plot(w_t, P_t, linewidth=2, color='black', alpha=0.4,
                label='P(w;x=L)')
        ax.plot(w0_t, P_t, linewidth=2, color='orange', alpha=0.4,
                label='P(w;x=0)')
        ax.set_ylim(ymin=ymin, ymax=ymax)
        ax.set_xlim(xmin=xmin, xmax=xmax)
        ax.set_ylabel('pull-out force P [N]')
        ax.set_xlabel('pull-out displacement [mm]')
        if self.show_legend:
            ax.legend(loc=4)
        self.plot_marker(ax, vot)

    def plot_marker(self, ax, vot):
        sim = self.vis2d.sim
        P_t, P0_t, w_t, w0_t = sim.hist['Pw'].Pw
        idx = sim.hist.get_time_idx(vot)
        P, P0, w, w0 = P_t[idx], P0_t[idx], w_t[idx], w0_t[idx]
        ax.plot([w], [P], 'o', color='black', markersize=10)
        ax.plot([w0], [P], 'o', color='orange', markersize=10)

    show_data = Button()

    def _show_data_fired(self):
        sim = self.vis2d.sim
        P_t, u_t = sim.hist['Pw'].Pw_right
        data = np.vstack([u_t, P_t]).T
        show_data = DataSheet(data=data)
        show_data.edit_traits()

    def plot_tex(self, ax, vot, *args, **kw):
        self.plot(ax, vot, *args, **kw)

    traits_view = View(
        Item('name', style='readonly'),
        Item('show_legend'),
        Item('show_data')
    )




from bmcs.time_functions import \
    LoadingScenario
from ibvpy.bcond import BCDof
from ibvpy.tmodel.mats3D.mats3D_plastic.vmats3D_desmorat import \
    MATS3DDesmorat
from simulator.api import TStepBC, XDomainSinglePoint

bc = BCDof(
    var='f', dof=0, value=-0.001,
    time_function=LoadingScenario()
)
model = TStepBC(
    domains=[(XDomainSinglePoint(), MATS3DDesmorat())],
    bc=[bc],
    debug=False
)
s = model.sim
s.tloop.verbose = True
s.run_thread()
s.join_thread()

print(model.hist.F_t)
print(model.hist.U_t)




from mayavi import mlab
import numpy as np


def decorate_figure(f, viz, distance=400, focal_point=(150, 40, 0)):
    mlab.view(0, 0, distance, np.array(focal_point), figure=f)
    mlab.orientation_axes(viz.src, figure=f)
    axes = mlab.axes(viz.src, figure=f)
    axes.label_text_property.trait_set(
        font_family='times', italic=False
    )
    axes.title_text_property.font_family = 'times'
    axes.axes.trait_set(
        x_label='x', y_label='y', z_label='z'
    )



'''
..todo:: compare with the current ibvpy.core.tloop and generalize the NRTLoop

..todo:: prepare the next example using sympy to allow for general functions

..todo:: make the `boundary condition' time dependent

..todo:: interface of a model - find an optimum handling of state transitions 
        with R, dR and state_update
        
..todo:: test the scenario - no convergence - and restart with modified tie step

..todo:: Handle the END event of the simulator in case 
that the calculation thread is still running. In such a case,
the request should be confirmed by the user action.

..todo:: Handle the resume state - should it recalculate the 
model staet after the pause event?


'''


'''This example couples two domains via 
an zero-thickness interface. 

Todo - simplify the domain-staet-xmodel-mesh hierarchy

The dependencies - Simulator - who sets the type of the time stepping
loop and the type of the time step.

Test two independent domains.

@todo:
- define the damage function with parameters realistic for concrete
- define the UI components representing the scale of the problem
- sort XDomains as mix-in class - define the class hierarchy and examples 
- reproduce the pullout from the fatigue paper
- test the the pressure sensitive bond
- include the microplane model with parameters calibrated on 3d bending
- instead of mlab get the bmcs window
- reproduce splitting
'''

import time

from ibvpy.bcond import BCSlice
from ibvpy.fets import FETS2D4Q
from ibvpy.fets.fets1D5.fets1d52ulrh import FETS1D52ULRH
from ibvpy.tmodel.mats1D5.vmats1D5_dp import \
    MATS1D5DP
from ibvpy.tmodel.mats3D.mats3D_plastic.vmats3D_desmorat import \
    MATS3DDesmorat
from ibvpy.tmodel.mats_damage_fn import \
    IDamageFn, LiDamageFn, JirasekDamageFn, AbaqusDamageFn, \
    MultilinearDamageFn, \
    FRPDamageFn
from ibvpy.tmodel.viz3d_scalar_field import \
    Vis3DStateField, Viz3DScalarField
from ibvpy.tmodel.viz3d_tensor_field import \
    Vis3DTensorField, Viz3DTensorField
from mayavi import mlab
from simulator.api import \
    TStepBC
from simulator.xdomain.xdomain_fe_grid_axisym import XDomainFEGridAxiSym
from simulator.xdomain.xdomain_interface import XDomainFEInterface

import numpy as np

from .mlab_decorators import decorate_figure


n_x_e = 30
n_y_e = 5
L_x = 300.0
R_in = 5.0
R_out = 75.0 / 2
xd1 = XDomainFEGridAxiSym(coord_min=(0, 0),
                          coord_max=(L_x, R_in),
                          shape=(n_x_e, 2),
                          fets=FETS2D4Q())
xd2 = XDomainFEGridAxiSym(coord_min=(0, R_in),
                          coord_max=(L_x, R_out),
                          shape=(n_x_e, n_y_e),
                          integ_factor=2 * np.pi,
                          fets=FETS2D4Q())
m1 = MATS3DDesmorat(E_1=210000, nu=0.3, tau_bar=2000.0)
m2 = MATS3DDesmorat(tau_bar=2.0)

xd12 = XDomainFEInterface(
    I=xd1.mesh.I[:, -1],
    J=xd2.mesh.I[:, 0],
    fets=FETS1D52ULRH()
)

u_max = 0.6 * 3
left_y = BCSlice(slice=xd1.mesh[0, 0, 0, 0],
                 var='u', dims=[1], value=0)
left_x = BCSlice(slice=xd1.mesh[0, :, 0, :],
                 var='u', dims=[0], value=-u_max)
right_x = BCSlice(slice=xd1.mesh[-1, :, -1, :],
                  var='u', dims=[0], value=u_max)
bc1 = [left_y, left_x, right_x]

m_interface = MATS1D5DP(E_N=1000, E_T=2000,
                        tau_bar=10)

m_interface.omega_fn = FRPDamageFn()
m = TStepBC(
    domains=[(xd1, m1),
             (xd2, m2),
             (xd12, m_interface),
             ],
    bc=bc1,  # + bc2,
)

m.hist.vis_record = {
    'strain': Vis3DTensorField(var='eps_ab'),
    'stress': Vis3DTensorField(var='sig_ab'),
    #  'damage': Vis3DStateField(var='omega_a'),
    #        'kinematic hardening': Vis3DStateField(var='z_a')
}

s = m.sim
s.tloop.k_max = 1000
s.tline.step = 0.01
s.tstep.fe_domain.serialized_subdomains
s.tloop.verbose = True

xd12.hidden = True
s.run_thread()
time.sleep(3)

mlab.options.backend = 'envisage'

f_strain = mlab.figure()
scene = mlab.get_engine().scenes[-1]
scene.name = 'stress'
strain_viz = Viz3DTensorField(vis3d=s.hist['strain'])
strain_viz.setup()
strain_viz.warp_vector.filter.scale_factor = 100.0
strain_viz.plot(s.tstep.t_n)

f_stress = mlab.figure()
scene = mlab.get_engine().scenes[-1]
scene.name = 'stress'
stress_viz = Viz3DTensorField(vis3d=s.hist['stress'])
stress_viz.setup()
stress_viz.warp_vector.filter.scale_factor = 100.0
stress_viz.plot(s.tstep.t_n)

if False:
    f_damage = mlab.figure()
    scene = mlab.get_engine().scenes[-1]
    scene.name = 'damage'
    damage_viz = Viz3DScalarField(vis3d=s.hist['damage'])
    damage_viz.setup()
    damage_viz.warp_vector.filter.scale_factor = 100.0
    damage_viz.plot(s.tstep.t_n)


decorate_figure(f_strain, strain_viz)
decorate_figure(f_stress, stress_viz)
# decorate_figure(f_damage, damage_viz)
mlab.show()




# Define the derivatives of the reference subspace
# with respect to the embedding space

from mathkit import EPS
import mayavi.mlab as mlab
import numpy as np
import sympy as sp
import traits.api as tr

eta_0, eta_1, eta_2 = sp.symbols('eta_0, eta_1, eta_2')


class Geo(tr.HasStrictTraits):

    eta = tr.List([eta_0, eta_1, eta_2])

    R_r = tr.List([sp.sin(eta_0) * eta_0, sp.cos(eta_0 + 1)**2, eta_1])

    def F_r(self, *eta_):
        r = np.array([
            re.subs({e: et for e, et in zip(self.eta, eta_)})
            for re in self.R_r
        ], dtype=np.float_)
        return np.einsum('ij,jk->ki', r[:, np.newaxis], np.identity(3))

    M_ra = tr.Property(depends_on='R_r, R_r_items')

    @tr.cached_property
    def _get_M_ra(self):
        return [[sp.diff(r, e) for r in self.R_r] for e in self.eta]

    def F_m(self, *eta_):
        return np.array(
            [
                [
                    m.subs({e: et for e, et in zip(self.eta, eta_)})
                    for m in m_a
                ]
                for m_a in self.M_ra
            ],
            dtype=np.float_
        )

    def orthogonal_base(self, *eta_):
        m_01 = self.F_m(*eta_)
        m_2 = np.einsum('...i,...j,ijk->...k', m_01[0, :], m_01[1, :], EPS)
        m_1a = np.einsum('...i,...j,ijk->...k', m_2, m_01[0, :], EPS)
        m = np.array([m_01[0, :], m_1a, m_2])
        m1 = m / np.sqrt(np.einsum('ij,ij->i', m, m))[:, np.newaxis]
        return m1

    def plot_basis(self, *eta):
        r = self.F_r(*eta)
        m = self.orthogonal_base(*eta)
        rm = r + m
        mp = np.vstack([r[np.newaxis, ...],
                        rm[np.newaxis, ...]])
        mlab.plot3d(mp[:, 0, 0], mp[:, 0, 1], mp[:, 0, 2],
                    color=black, tube_radius=0.01)
        mlab.plot3d(mp[:, 1, 0], mp[:, 1, 1], mp[:, 1, 2],
                    color=black, tube_radius=0.01)
        mlab.plot3d(mp[:, 2, 0], mp[:, 2, 1], mp[:, 2, 2],
                    color=black, tube_radius=0.01)
        mlab.text3d(rm[0, 0], rm[0, 1], rm[0, 2],
                    'X', color=black, scale=0.1)
        mlab.text3d(rm[1, 0], rm[1, 1], rm[1, 2],
                    'Y', color=black, scale=0.1)
        mlab.text3d(rm[2, 0], rm[2, 1], rm[2, 2],
                    'Z', color=black, scale=0.1)


g = Geo()

g.R_r = [eta_0, eta_1 * eta_2, eta_0 * eta_1]
g.R_r = [sp.sin(eta_0) * eta_0, sp.cos(eta_0 + eta_1 + 1)**2, eta_1 + eta_2]
#g.R_r = [sp.sin(eta_0), sp.cos(eta_0), eta_1]

black = (0, 0, 0)
white = (1, 1, 1)

ee0, ee1, ee2 = np.mgrid[0:2:10j, 0:1:5j, 0:1:5j]

mlab.figure(bgcolor=white)
for e0, e1, e2 in zip(ee0.flatten(), ee1.flatten(), ee2.flatten()):
    g.plot_basis(e0, e1, e2)

mlab.show()
